#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Standard imports
import base64
import logging
import pprint
from io import BytesIO
from pathlib import Path

import sys
import os

if sys.version_info.major >= 3:
    from html import escape
else:
    print("ExportPDF running with Python < 3 : stability and output not guaranteed. Please run exportPDF with at least Python3")

logger = logging.getLogger('pymisp')

# Potentially not installed imports
try:
    from reportlab.pdfgen import canvas  # type: ignore
    from reportlab.pdfbase.pdfmetrics import stringWidth, registerFont  # type: ignore
    from reportlab.pdfbase.ttfonts import TTFont  # type: ignore
    from reportlab.lib import colors  # type: ignore
    from reportlab.lib.pagesizes import A4  # type: ignore

    from reportlab.platypus import SimpleDocTemplate, Paragraph, PageBreak, Table, TableStyle, Flowable, Image, Indenter  # type: ignore

    from reportlab.lib.styles import getSampleStyleSheet, ParagraphStyle  # type: ignore
    from reportlab.lib.units import mm  # type: ignore
    from reportlab.lib.enums import TA_CENTER, TA_JUSTIFY, TA_LEFT  # type: ignore

    HAS_REPORTLAB = True
except ImportError:
    HAS_REPORTLAB = False


########################################################################
def create_flowable_tag(misp_tag):
    '''
    Returns a Flowable tag linked to one tag.
    :param misp_tag: A misp tag of a misp event or a misp event's attribute
    :return: one flowable representing a tag (with style)
    '''
    col1_style, col2_style = get_table_styles()

    return [Flowable_Tag(text=misp_tag.name, color=misp_tag.colour, custom_style=col1_style)]


class Flowable_Tag(Flowable):
    """
    Custom flowable to handle tags. Draw one Tag with the webview formatting
    Modified from : http://two.pairlist.net/pipermail/reportlab-users/2005-February/003695.html
    and : http://www.blog.pythonlibrary.org/2014/03/10/reportlab-how-to-create-custom-flowables/
    """

    # ----------------------------------------------------------------------
    def __init__(self, x=0, y=0, width=40, height=15, text="", color="#ffffff", custom_style=None):
        Flowable.__init__(self)
        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.text = text
        self.colour = color
        if custom_style is not None:
            self.custom_style = custom_style
        else:
            self.custom_style = getSampleStyleSheet()["Normal"]

    # ----------------------------------------------------------------------

    def coord(self, x, y, unit=1):
        """
        http://stackoverflow.com/questions/4726011/wrap-text-in-a-table-reportlab
        Helper class to help position flowables in Canvas objects
        """
        x, y = x * unit, self.height - y * unit
        return x, y

    # ----------------------------------------------------------------------
    def __repr__(self):
        return "Tag(w=" + str(self.width) + ")"

    # ----------------------------------------------------------------------
    def choose_good_text_color(self):
        # See : http://trendct.org/2016/01/22/how-to-choose-a-label-color-to-contrast-with-background/
        r, g, b = colors.HexColor(self.colour).rgb()

        brightness = r * 299 + g * 587 + b * 114 / 1000

        if brightness < 500:  # Standard treeshold for human vision : 123 instead of 500
            text_color = "#ffffff"  # Black
        else:
            text_color = "#000000"  # White

        return text_color

    # ----------------------------------------------------------------------
    def draw(self):
        '''
        Draw the shape, text, etc to show a Tag
        Honestely, constant are totally ad-hoc. Feels free to change it, but be sure to test the visual result of it.
        '''
        RADIUS = 1 * mm
        LEFT_INTERNAL_PADDING = 2
        ELONGATION = LEFT_INTERNAL_PADDING * 2

        p = Paragraph("<font color='{}'>{}</font>".format(self.choose_good_text_color(), self.text), style=self.custom_style)
        string_width = stringWidth(self.text, self.custom_style.fontName, self.custom_style.fontSize)

        self.width = string_width + ELONGATION
        self.height = self.custom_style.fontSize

        self.canv.setFillColor(colors.HexColor(self.colour))
        self.canv.roundRect(self.x, self.y + LEFT_INTERNAL_PADDING, self.width, self.height + 2, RADIUS, fill=1)

        p.wrapOn(self.canv, self.width, self.height)
        p.drawOn(self.canv, *self.coord(self.x, self.y + 0.5 * LEFT_INTERNAL_PADDING, mm))


# Copy of pdfexport.py moduleconfig
moduleconfig = ["MISP_base_url_for_dynamic_link", "MISP_name_for_metadata", "Activate_textual_description",
                "Activate_galaxy_description", "Activate_related_events", "Activate_internationalization_fonts",
                "Custom_fonts_path"]

# == Row colors of the table (alternating) ==
EVEN_COLOR = colors.whitesmoke
ODD_COLOR = colors.lightgrey
EVEN_COLOR_GALAXY = colors.powderblue
ODD_COLOR_GALAXY = colors.lavenderblush

# == Lines parameters of the table ==
LINE_COLOR = colors.lightslategray
LINE_THICKNESS = 0.75

# == Columns colors, aligment, fonts, space, size, width, heights ==
FIRST_COL_FONT_COLOR = colors.HexColor("#333333")  # Same as GUI
FIRST_COL_FONT = 'Helvetica-Bold'
FIRST_COL_ALIGNEMENT = TA_CENTER

SECOND_COL_FONT_COLOR = colors.black
SECOND_COL_FONT = 'Helvetica'
SECOND_COL_ALIGNEMENT = TA_LEFT

TEXT_FONT_SIZE = 8
LEADING_SPACE = 7

# Small clusters fonts
SMALL_FONT_SIZE = TEXT_FONT_SIZE - 1
SMALL_LEADING_SPACE = LEADING_SPACE
SMALL_COL1_ALIGMENT = FIRST_COL_ALIGNEMENT
SMALL_COL2_ALIGMENT = TA_JUSTIFY

EXPORT_DATE_FORMAT = '%Y-%m-%d %H:%M:%S'
COL_WIDTHS = ['25%', '75%']  # colWidths='*' # Not documented but does exist
ROW_HEIGHT = 5 * mm  # 4.5 * mm (a bit too short to allow vertical align TODO : Fix it)
ROW_HEIGHT_FOR_TAGS = 4 * mm  # 4.5 * mm (a bit too short to allow vertical align TODO : Fix it)

# == Whole document margins and size ==
PAGESIZE = A4  # (140 * mm, 216 * mm)  # width, height
BASE_MARGIN = 5 * mm  # Create a list here to specify each row separately
INDENT_SIZE = 6 * mm  # The Indentation of attribute from object, etc.
INDENT_SIZE_HEADING = 3 * mm  # The Indentation of attribute from object, etc.

# == Parameters for error handling for content too long to fit on a page ==
FRAME_MAX_HEIGHT = 200 * mm  # 500  # Ad hoc value for a A4 page
FRAME_MAX_WIDTH = 145 * mm - INDENT_SIZE  # 356
STR_TOO_LONG_WARNING = "<br/><b><font color=red>[Too long to fit on a single page. Cropped]</font></b>"

# == Parameters for error handling for image too big to fit on a page ==
FRAME_PICTURE_MAX_HEIGHT = 200 * mm  # 195 * mm
FRAME_PICTURE_MAX_WIDTH = 145 * mm - INDENT_SIZE  # 88 * mm

# == Parameters for links management ==
LINK_TYPE = "link"  # Name of the type that define 'good' links
URL_TYPE = "url"  # Name of the type that define 'bad' links
IMAGE_TYPE = "attachment"  # /!\ Not only pictures ! Can be PDF, ...
WARNING_MESSAGE_URL = "'https://Please_consider_that_this_may_be_a_harmful_link'"
NOT_A_PICTURE_MESSAGE = "This attachment is not recognized as an image. Please access this attachment directly from your MISP instance."
GOOD_LINK_COLOR = 'blue'
BAD_LINK_COLOR = 'red'

# == Parameters for description ==
LOW_THREAT_COLOR = 'green'
MEDIUM_THREAT_COLOR = 'orange'
HIGH_THREAT_COLOR = 'red'
EXTERNAL_ANALYSIS_PREFIX = "<i>External analysis from an attribute : </i>"

DEFAULT_VALUE = "No value specified."

# == Parameters for improvement of event's metadata ==

threat_map = {"0": f"<font color ={MEDIUM_THREAT_COLOR}>   undefined (0)</font>",
              "3": f"<font color ={LOW_THREAT_COLOR}>      Low (3)</font>",
              "2": f"<font color ={MEDIUM_THREAT_COLOR}>   Medium (2)</font>",
              "1": f"<font color ={HIGH_THREAT_COLOR}>     High (1)</font>"}

analysis_map = {"0": f"<font color ={HIGH_THREAT_COLOR}>   Initial (0)</font>",
                "1": f"<font color ={MEDIUM_THREAT_COLOR}> Ongoing (1)</font>",
                "2": f"<font color ={LOW_THREAT_COLOR}>    Completed (2)</font>"}

# == Parameters for Sightings ==
POSITIVE_SIGHT_COLOR = 'green'
NEGATIVE_SIGHT_COLOR = 'red'
MISC_SIGHT_COLOR = 'orange'

# == Parameters for galaxies ==
DO_SMALL_GALAXIES = True
FIRST_LEVEL_GALAXY_WIDTHS = ["15%", "85%"]
SECOND_LEVEL_GALAXY_WIDTHS = ["20%", "80%"]
CLUSTER_COLORS = [0]  # or 1
OFFSET = 1

# == Parameters of published value ==
RED_COLOR = '#ff0000'
GREEN_COLOR = '#008000'
YES_ANSWER = f"<font color={GREEN_COLOR}><b> Yes </b></font>"
NO_ANSWER = f"<font color={RED_COLOR}><b> No </b></font>"


########################################################################
# "UTILITIES" METHODS. Not meant to be used except for development purposes

def get_sample_fonts():
    '''
    Get fonts available on the current system, usable in pdf generation
    :return: None. Print on std output the list of available fonts
    '''

    # Create a dummy canvas
    c = canvas.Canvas("hello.pdf")

    # Print list of usable fonts
    pprint.pprint(c.getAvailableFonts())


def get_sample_styles():
    '''
    Get styles available in reportLab (Paragraph, Heading1, ...)
    :return: None. Print on std output the list of available styles
    '''

    # Get styles, as for example sample_style_sheet['Heading1'], sample_style_sheet['BodyText'] ...
    sample_style_sheet = getSampleStyleSheet()

    # if you want to see all the sample styles, this prints them
    sample_style_sheet.list()


########################################################################
# General Event's Attributes formater tools

def uuid_to_url(baseurl, uuid):
    '''
    Return an url constructed from the MISP baseurl and the uuid of the event, to go to this event on this MISP
    :param baseurl: the baseurl of the MISP instnce e.g. http://localhost:8080 or http://localhost:8080/
    :param uuid: the uuid of the event that we want to have a link to
    :return: the complete URL to go to this event on this MISP instance
    '''
    if baseurl[len(baseurl) - 1] != "/":
        baseurl += "/"
    return f"{baseurl}events/view/{uuid}"


def create_flowable_table_from_data(data, col_w=COL_WIDTHS, color_alternation=None,
                                    line_alternation=None, galaxy_colors=False):
    '''
    Given a list of flowables items (2D/list of list), creates a Table with styles.
    :param data: list of list of items (flowables is better)
    :return: a Table - with styles - to add to the pdf
    '''
    # Create the table
    curr_table = Table(data, col_w)

    # Aside notes :
    #   colWidths='*' does a 100% and share the space automatically
    #   rowHeights=ROW_HEIGHT if you want a fixed height. /!\ Problems with paragraphs that are spreading everywhere

    # Create styles and set parameters
    alternate_colors_style = alternate_colors_style_generator(data, color_alternation, galaxy_colors)
    lines_style = lines_style_generator(data, line_alternation)
    general_style = general_style_generator()

    # Make the table nicer
    curr_table.setStyle(TableStyle(general_style + alternate_colors_style + lines_style))

    return curr_table


def alternate_colors_style_generator(data, color_alternation, galaxy_colors=True):
    '''
    Create a style, applicable on a table that will be built with parameter's data, with alternated
    background color for each line.
    Modified from : https://gist.github.com/chadcooper/5798392
    :param color_alternation: Allow to control the color scheme. e.g. [0,0,0,1,1,0 ... will produce 3 lines of a color,
    2 lines of another, 1 of the first one ...
    :param data: list of list of items (2D table) to be displayed in the pdf
    :return: A list of 'BACKGROUND' properties, usable in a TableStyle, with alternated colours
    '''

    data_len = len(data)
    color_list = []

    if color_alternation is None:
        # For each line, generate a tuple giving to a line a color
        for each in range(data_len):
            if each % 2 == 0:
                bg_color = EVEN_COLOR if not galaxy_colors else EVEN_COLOR_GALAXY
            else:
                bg_color = ODD_COLOR if not galaxy_colors else ODD_COLOR_GALAXY
            color_list.append(('BACKGROUND', (0, each), (-1, each), bg_color))
    else:
        # if data_len > len(color_alternation) :
        #    logger.warning("Line alternation for PDF display isn't correctly set. Looping on given values only.")

        # For each line, generate a tuple giving to a line a color
        for each in range(data_len):
            if color_alternation[each % len(color_alternation)] % 2 == 0:
                bg_color = EVEN_COLOR if not galaxy_colors else EVEN_COLOR_GALAXY
            else:
                bg_color = ODD_COLOR if not galaxy_colors else ODD_COLOR_GALAXY
            color_list.append(('BACKGROUND', (0, each), (-1, each), bg_color))

    return color_list


def lines_style_generator(data, line_alternation):
    '''
    Create a style, applicable on a table that will be built with parameter's data,
    that draw colored lines above and below each line of the table
    :param line_alternation: Allow to control the color scheme. e.g. [0,0,0,1,1,0 ... will produce with a line up it,
    2 lines without, 1 of the first one ...
    :param data:  list of list of items (2D table) to be displayed in the pdf
    :return: A list of 'LINE****' properties, usable in a TableStyle, that are drawing lines
    '''
    data_len = len(data)
    lines_list = []

    if line_alternation is None:
        # For each line, generate a tuple giving to a line a color
        for each in range(data_len):
            lines_list.append(('LINEABOVE', (0, each), (-1, each), LINE_THICKNESS, LINE_COLOR))

        # Last line
        lines_list.append(('LINEBELOW', (0, len(data) - 1), (-1, len(data) - 1), LINE_THICKNESS, LINE_COLOR))
    elif line_alternation == []:
        # Do nothing
        return lines_list
    else:
        # if data_len > len(line_alternation) :
        #    logger.warning("Line alternation for PDF display isn't correctly set. Looping on given values only.")

        # For each line, generate a tuple giving to a line a color
        for each in range(data_len):
            if each == 0 or line_alternation[each % len(line_alternation)] != line_alternation[(each - 1) % len(line_alternation)]:
                lines_list.append(('LINEABOVE', (0, each), (-1, each), LINE_THICKNESS, LINE_COLOR))

        # Last line
        lines_list.append(('LINEBELOW', (0, len(data) - 1), (-1, len(data) - 1), LINE_THICKNESS, LINE_COLOR))

    return lines_list


def general_style_generator():
    '''
    Create the general style (alignement, padding ...) of the table, copying the MISP'event's web_view.
    :return: a list of properties, usable in a TableStyle
    '''
    lines_list = []

    lines_list.append(('VALIGN', (0, 0), (-1, -1), 'MIDDLE'))
    lines_list.append(('LEFTPADDING', (0, 0), (-1, -1), 0))
    lines_list.append(('RIGHTPADDING', (0, 0), (-1, -1), 0))

    # VERTICAL_PADDING = 2
    # lines_list.append(('TOPPADDING', (0, 0), (-1, -1), VERTICAL_PADDING))
    # lines_list.append(('BOTTOMPADDING', (0, 0), (-1, -1), VERTICAL_PADDING))

    return lines_list


def internationalize_font(config=None):
    global FIRST_COL_FONT
    global SECOND_COL_FONT

    if is_in_config(config, 6) and config[moduleconfig[6]] != "":
        # Handle custom fonts. Has to be TrueType = one TTF only
        fonts_path_custom = config[moduleconfig[6]]

        if fonts_path_custom and os.path.isfile(fonts_path_custom):
            registerFont(TTFont("custom_font", fonts_path_custom))
            FIRST_COL_FONT = 'custom_font'
            SECOND_COL_FONT = 'custom_font'

        else:
            logger.error(f"Trying to load a custom font, unable to access the file. Path: {fonts_path_custom}")
    else:
        ''' Handle provided NOTO fonts (CJK only for now)
            # Available fonts :
            NotoSansCJKtc - DemiLight.ttf
            NotoSansCJKtc - Regular.ttf
            NotoSansCJKtc - Black.ttf
            NotoSansCJKtc - Light.ttf
            NotoSansCJKtc - Thin.ttf
            NotoSansCJKtc - Bold.ttf
            NotoSansCJKtc - Medium.ttf
        '''
        font_path = Path(sys.modules['pymisp'].__file__).parent / 'tools' / 'pdf_fonts' / 'Noto_TTF'

        noto_bold = font_path / "NotoSansCJKtc-Bold.ttf"
        noto = font_path / "NotoSansCJKtc-DemiLight.ttf"

        if noto_bold.is_file() and noto.is_file():
            registerFont(TTFont("Noto", noto))
            registerFont(TTFont("Noto-bold", noto_bold))

            FIRST_COL_FONT = 'Noto-bold'
            SECOND_COL_FONT = 'Noto'
        else:
            logger.error(f"Trying to load a custom (internationalization) font, unable to access the file: {noto_bold}")


def get_table_styles():
    '''
    Create and returns the two mains styles for the columns of the document.
    :return: two styles, one for each columns of the document, describing the MISP object.
    '''
    sample_style_sheet = getSampleStyleSheet()

    custom_body_style_col_1 = ParagraphStyle(name='Column_1',
                                             parent=sample_style_sheet['Normal'],
                                             fontName=FIRST_COL_FONT,
                                             textColor=FIRST_COL_FONT_COLOR,
                                             fontSize=TEXT_FONT_SIZE,
                                             leading=LEADING_SPACE,
                                             alignment=FIRST_COL_ALIGNEMENT)

    custom_body_style_col_2 = ParagraphStyle(name='Column_2',
                                             parent=sample_style_sheet['Normal'],
                                             fontName=SECOND_COL_FONT,
                                             textColor=SECOND_COL_FONT_COLOR,
                                             fontSize=TEXT_FONT_SIZE,
                                             leading=LEADING_SPACE,
                                             alignment=SECOND_COL_ALIGNEMENT)

    return custom_body_style_col_1, custom_body_style_col_2


def get_clusters_table_styles():
    '''
    Create and returns the two mains styles for the columns of a table describing a cluster.
    :return: two styles, one for each columns of the document, describing the MISP object.
    '''
    col1, col2 = get_table_styles()

    custom_body_style_col_1 = ParagraphStyle(name='Column_1_small',
                                             parent=col1,
                                             fontName=FIRST_COL_FONT,
                                             textColor=FIRST_COL_FONT_COLOR,
                                             fontSize=SMALL_FONT_SIZE,
                                             leading=SMALL_LEADING_SPACE,
                                             alignment=SMALL_COL1_ALIGMENT)

    custom_body_style_col_2 = ParagraphStyle(name='Column_2_small',
                                             parent=col2,
                                             fontName=SECOND_COL_FONT,
                                             textColor=SECOND_COL_FONT_COLOR,
                                             fontSize=SMALL_FONT_SIZE,
                                             leading=SMALL_LEADING_SPACE,
                                             alignment=SMALL_COL2_ALIGMENT)

    return custom_body_style_col_1, custom_body_style_col_2


########################################################################
# Checks

def safe_string(bad_str):
    return escape(str(bad_str))


def is_safe_value(value):
    return (value is not None
            and value != "")


def is_safe_table(value):
    return (value is not None
            and value != [])


def is_safe_attribute(curr_object, attribute_name):
    return (hasattr(curr_object, attribute_name)
            and getattr(curr_object, attribute_name) is not None
            and getattr(curr_object, attribute_name) != "")


def is_safe_dict_attribute(curr_object, attribute_name):
    return (attribute_name in curr_object
            and curr_object[attribute_name] is not None
            and curr_object[attribute_name] != "")


def is_safe_attribute_table(curr_object, attribute_name):
    return (hasattr(curr_object, attribute_name)
            and getattr(curr_object, attribute_name) is not None
            and getattr(curr_object, attribute_name) != [])


def is_in_config(config, index):
    return config is not None and moduleconfig[index] in config


########################################################################
# Functions grouped by misp object type


class Value_Formatter():
    '''
    "item" parameter should be as follow, a list of name, in order :
        ["Name to be print in the pdf", "json property access name",
        " Name to be display if no values found in the misp_event"]
    '''

    # ----------------------------------------------------------------------
    def __init__(self, config, col1_style, col2_style, col1_small_style, col2_small_style):
        self.config = config
        self.col1_style = col1_style
        self.col2_style = col2_style
        self.col1_small_style = col1_small_style
        self.col2_small_style = col2_small_style

    # ----------------------------------------------------------------------
    ########################################################################
    # General attribut formater
    def get_col1_paragraph(self, dirty_string, do_small=False):
        if do_small:
            return self.get_unoverflowable_paragraph(dirty_string, self.col1_small_style, do_small=do_small)
        return self.get_unoverflowable_paragraph(dirty_string, self.col1_style, do_small=do_small)

    def get_unoverflowable_paragraph(self, dirty_string, curr_style=None, do_escape_string=True, do_small=False):
        '''
        Create a paragraph that can fit on a cell displayed one page maximum.
        This method can be improved (get the exact size of the current frame, and limit the paragraph to this size.)
        KeepInFrame may give a nicer solution (not for me so far)
        :param do_escape_string: Activate the escaping (may be useful to add inline HTML, e.g. hyperlinks)
        :param dirty_string: String to transform
        :param curr_style: Style to apply to the returned paragraph
        :return:
        '''
        if do_escape_string:
            sanitized_str = safe_string(dirty_string)
        else:
            sanitized_str = dirty_string

        if curr_style is None:
            if do_small:
                curr_style = self.col2_small_style
            else:
                curr_style = self.col2_style

        # Get the space that the paragraph needs to be printed
        w, h = Paragraph(sanitized_str, curr_style).wrap(FRAME_MAX_WIDTH, FRAME_MAX_HEIGHT)

        # If there is enough space, directly send back the sanitized paragraph
        if w <= FRAME_MAX_WIDTH and h <= FRAME_MAX_HEIGHT:
            answer_paragraph = Paragraph(sanitized_str, curr_style)
        else:
            # Otherwise, cut the content to fit the paragraph (Dichotomy)
            max_carac_amount = int((FRAME_MAX_HEIGHT / (h * 1.0)) * len(sanitized_str))

            i = 0
            MAX_ITERATION = 10
            limited_string = ""
            while (w > FRAME_MAX_WIDTH or h > FRAME_MAX_HEIGHT) and i < MAX_ITERATION:
                i += 1
                limited_string = sanitized_str[:max_carac_amount]  # .replace("\n", "").replace("\r", "")
                w, h = Paragraph(limited_string + STR_TOO_LONG_WARNING, curr_style).wrap(FRAME_MAX_WIDTH,
                                                                                         FRAME_MAX_HEIGHT)
                max_carac_amount = int(max_carac_amount / 2)

            if w <= FRAME_MAX_WIDTH and h <= FRAME_MAX_HEIGHT:
                answer_paragraph = Paragraph(limited_string + STR_TOO_LONG_WARNING, curr_style)
            else:
                # We may still end with a not short enough string
                answer_paragraph = Paragraph(STR_TOO_LONG_WARNING, curr_style)

        return answer_paragraph

    def get_value_link_to_event(self, uuid=None, text=None, curr_style=None, color=True):
        '''
        Returns a flowable paragraph to add to the pdf given the misp_event uuid, with or without link
        :param curr_style: style to apply to the returned paragraph
        :param text: text to which the link will be anchored
        :param uuid: used to construct the link
        :param color: Boolean to give a color or not to the generate link (good link color)
        :return: a Paragraph to add in the pdf, regarding the values of "uuid"
        '''

        if curr_style is None:
            curr_style = self.col2_style

        escape = True
        # Does MispEvent has the attribute ?
        if is_safe_value(text):
            # It has the requested attribute .. building upon it.

            # Does misp_object has an uuid and do we know the baseurl ?
            if is_safe_value(uuid) and is_in_config(self.config, 0):
                # We can build links
                escape = False
                curr_uuid = str(is_safe_value(uuid))
                curr_baseurl = self.config[moduleconfig[0]]
                curr_url = uuid_to_url(curr_baseurl, curr_uuid)
                html_url = "<a href={}>{}</a>".format(curr_url, safe_string(text))

                if color:
                    # They want fancy colors
                    html_url = f"<font color={GOOD_LINK_COLOR}>{html_url}</font>"
                # Construct final paragraph
                answer = html_url

            else:
                # We can't build links
                answer = text

        else:
            # No it doesn't, so we directly give the default answer
            answer = DEFAULT_VALUE

        if not escape:
            return self.get_unoverflowable_paragraph(answer, curr_style=curr_style, do_escape_string=False)
        return self.get_unoverflowable_paragraph(answer, curr_style=curr_style)

    ########################################################################
    # Specific attribute formater

    def get_date_value(self, date=None):
        '''
        Returns a flowable paragraph to add to the pdf given the misp_event date
        :param date: MISP_EVENT date to be formatted
        :return: a Paragraph to add in the pdf, regarding the values of "date"
        '''
        answer = DEFAULT_VALUE
        if is_safe_value(date):
            answer = safe_string(date)

        return self.get_unoverflowable_paragraph(answer)

    def get_owner_value(self, owner=None):
        '''
        Returns a flowable paragraph to add to the pdf given the misp_event owner
        :param owner: MISP_EVENT owner to be formatted
        :return: a Paragraph to add in the pdf, regarding the values of "owner"
        '''
        answer = DEFAULT_VALUE

        if is_safe_value(owner):
            answer = safe_string(owner)

        return self.get_unoverflowable_paragraph(answer)

    def get_threat_value(self, threat_level=None):
        '''
        Returns a flowable paragraph to add to the pdf given the misp_event threat
        :param threat_level: MISP_EVENT threat level (int) to be formatted
        :return: a Paragraph to add in the pdf, regarding the values of "threat"
        '''
        answer = "No threat level specified."

        if is_safe_value(threat_level) and str(threat_level) in threat_map:
            answer = threat_map[safe_string(threat_level)]

        return self.get_unoverflowable_paragraph(answer, do_escape_string=False)

    def get_analysis_value(self, analysis_level=None):
        '''
        Returns a flowable paragraph to add to the pdf given the misp_event analysis
        :param analysis_level: MISP_EVENT analysis level (int) to be formatted
        :return: a Paragraph to add in the pdf, regarding the values of "analysis"
        '''
        answer = "No analysis status specified."

        if is_safe_value(analysis_level) and str(analysis_level) in analysis_map:
            answer = analysis_map[safe_string(analysis_level)]

        return self.get_unoverflowable_paragraph(answer, do_escape_string=False)

    def get_timestamp_value(self, timestamp=None):
        '''
        Returns a flowable paragraph to add to the pdf given the misp_event timestamp
        :param timestamp: MISP_EVENT timestamp (int) to be formatted
        :return: a Paragraph to add in the pdf, regarding the values of "timestamp"
        '''
        answer = "No timestamp specified."

        if is_safe_value(timestamp):
            answer = safe_string(timestamp.strftime(EXPORT_DATE_FORMAT))

        return self.get_unoverflowable_paragraph(answer)

    def get_creator_organisation_value(self, creator=None):
        '''
        Returns a flowable paragraph to add to the pdf given the misp_event creator organisation
        :param creator: MISP_EVENT creator (not the name directly) to be formatted
        :return: a Paragraph to add in the pdf, regarding the values of "creator organisation"
        '''
        answer = DEFAULT_VALUE

        if is_safe_value(creator) and is_safe_value(creator.get('name', None)):
            answer = safe_string(creator.get('name', None))

        return self.get_unoverflowable_paragraph(answer)

    def get_attributes_number_value(self, attributes=None):
        '''
        Returns a flowable paragraph to add to the pdf given the misp_event attributes
        :param attributes: MISP_EVENT attributes list to be formatted
        :return: a Paragraph to add in the pdf, regarding the values of "attributes"
        '''
        answer = "0 - no attribute"

        if is_safe_table(attributes):
            answer = safe_string(len(attributes))

        return self.get_unoverflowable_paragraph(answer)

    def get_published_value(self, published_bool=None, published_timestamp=None):
        '''
        Returns a flowable paragraph to add to the pdf given the misp_event published/published_time
        More information on how to play with paragraph into reportlab cells :
        https://stackoverflow.com/questions/11810008/reportlab-add-two-paragraphs-into-one-table-cell
        :param published_timestamp: MISP_EVENT published_timestamp to be formatted
        :param published_bool: MISP_EVENT published boolean value
        :return: a Paragraph to add in the pdf, regarding the values of "published"/"publish_timestamp"
        '''

        # Formatting similar to MISP Event web view
        if is_safe_value(published_bool):
            if published_bool:  # == True
                answer = YES_ANSWER
                if is_safe_value(published_timestamp):
                    # Published and have published date
                    answer += '({})'.format(published_timestamp.strftime(EXPORT_DATE_FORMAT))
                else:
                    # Published without published date
                    answer += "(no date)"

            else:
                # Not published
                answer = NO_ANSWER
        else:
            # Does not have a published attribute
            answer = DEFAULT_VALUE

        return self.get_unoverflowable_paragraph(answer, do_escape_string=False)

    def get_image_value(self, image_buffer=None):
        '''
        Returns a flowable image to add to the pdf given the misp attribute type and data
        :param image_buffer: an image contained an attribute, for example (buffer / Base 64)
        :return: a flowable image to add in the pdf, regarding the values of "data"
        '''

        try:
            # Get the image
            buf = image_buffer  # TODO : Do verification on the buffer ?

            # Create image within a bounded box (to allow pdf creation)
            img = Image(buf, width=FRAME_PICTURE_MAX_WIDTH, height=FRAME_PICTURE_MAX_HEIGHT, kind='bound')
            answer = img

        except OSError:
            logger.error("Trying to add an attachment during PDF export generation. Attachement joining failed. Attachement may not be an image.")
            answer = self.get_unoverflowable_paragraph(f"<font color={BAD_LINK_COLOR}>{NOT_A_PICTURE_MESSAGE}</font>", do_escape_string=False)

        return answer

    def get_good_link(self, value=None):
        '''
        Returns a flowable paragraph to add to the pdf given the misp_attribute value, if this is a link
        :param value: string, of an url to format as a good link
        :return: a Paragraph to add in the pdf, regarding the values of this "link" attribute
        '''
        return self.get_unoverflowable_paragraph(f"<font color={GOOD_LINK_COLOR}><a href={value}>{value}</a></font>", do_escape_string=False)

    def get_bad_link(self, value=None):
        '''
        Returns a flowable paragraph to add to the pdf given the misp_attribute value, if this is a link
        :param value: string, of an url to format as a bad link
        :return: a Paragraph to add in the pdf, regarding the values of this "url" attribute
        '''
        return self.get_unoverflowable_paragraph(f"<font color={BAD_LINK_COLOR}><a href={WARNING_MESSAGE_URL}>{value}</a></font>", do_escape_string=False)

    def get_good_or_bad_link(self, value=None, type=None):
        '''
        Returns a flowable paragraph to add to the pdf given the misp_attribute value, if this is a link or an url
        :param type: Type of the url (url or link) as a string
        :param value: string, of an url to format as a good or bad link
        :return: a Paragraph to add in the pdf, regarding the values of this "link" or "url" attribute
        '''

        answer = self.get_unoverflowable_paragraph("Not an URL")

        # Handle "Good" links
        if type == LINK_TYPE:
            answer = self.get_good_link(value=value)
        # Handle "bad "links
        elif type == URL_TYPE:
            answer = self.get_bad_link(value=value)

        return answer

    def get_galaxy_name_value(self, misp_galaxy):
        '''
        Create a displayable name for a galaxy
        :param misp_galaxy: MISP_EVENT galaxy, as an object (not a list)
        :return: a Flowable Paragraph to add in the pdf, regarding the value of the MISP galaxy
        '''
        answer = DEFAULT_VALUE

        if is_safe_dict_attribute(misp_galaxy, 'name'):
            answer = '{} <i>from</i> {}:{}'.format(safe_string(misp_galaxy['name']),
                                                   safe_string(misp_galaxy["namespace"]),
                                                   safe_string(misp_galaxy["type"]))

        return self.get_unoverflowable_paragraph(answer, do_small=True)

    def get_galaxy_cluster_name_value(self, misp_cluster, do_small=False):
        '''
        Create a displayable name for a cluster
        :param misp_cluster: a MISP_EVENT's GALAXY's cluster, as an object (not a list)
        :param do_small: Compress the display (reduce the size of the flowable paragraph, fonts, etc.)
        :return: a Flowable Paragraph to add in the pdf, regarding the value of the MISP cluster
        '''
        # TODO : To be changed when Clust becomes an object
        tmp_text = ""

        if is_safe_dict_attribute(misp_cluster, 'value'):
            tmp_text += safe_string(misp_cluster['value'])

            # if is_safe_dict_attribute(misp_cluster, item[3]) :
            # tmp_text += "<br/><i>Source :</i> " + misp_cluster[item[3]]

            if is_safe_dict_attribute(misp_cluster, "meta") and is_safe_dict_attribute(misp_cluster["meta"], "synonyms"):
                tmp_text += " <br/><i>Synonyms :</i> "
                for i, synonyme in enumerate(misp_cluster["meta"]["synonyms"]):
                    if i != 0:
                        tmp_text += " / "
                    tmp_text += safe_string(synonyme)

            return self.get_unoverflowable_paragraph(tmp_text, do_escape_string=False, do_small=do_small)
        return self.get_unoverflowable_paragraph(DEFAULT_VALUE, do_small=do_small)


class Event_Metadata():

    # ----------------------------------------------------------------------
    def __init__(self, config, value_formatter):
        self.config = config
        self.value_formatter = value_formatter
        self.sample_style_sheet = getSampleStyleSheet()

    # ----------------------------------------------------------------------

    ########################################################################
    # General Event's Attributes formater

    def create_flowable_table_from_event(self, misp_event):
        '''
        Returns Table presenting a MISP event
        :param misp_event: A misp event (complete or not)
        :return: a table that can be added to a pdf
        '''

        data = []
        flowable_table = []

        # Manual addition
        # UUID
        data.append([self.value_formatter.get_col1_paragraph("UUID"),
                     self.value_formatter.get_value_link_to_event(uuid=misp_event.get('uuid', None),
                                                                  text=misp_event.get('uuid', None))])

        # Date
        data.append({self.value_formatter.get_col1_paragraph("Date"),
                     self.value_formatter.get_date_value(date=misp_event.get('date', None))})

        # Owner
        data.append([self.value_formatter.get_col1_paragraph("Owner org"),
                     self.value_formatter.get_owner_value(owner=misp_event.get('owner', None))])

        # Threat
        data.append([self.value_formatter.get_col1_paragraph("Threat level"),
                     self.value_formatter.get_threat_value(threat_level=misp_event.get('threat_level_id', None))])

        # Analysis
        data.append([self.value_formatter.get_col1_paragraph("Analysis"),
                     self.value_formatter.get_analysis_value(analysis_level=misp_event.get('analysis', None))])

        # Info
        data.append([self.value_formatter.get_col1_paragraph("Info"),
                     self.value_formatter.get_value_link_to_event(uuid=misp_event.get('uuid', None),
                                                                  text=misp_event.get('info', None))])

        # Timestamp
        data.append([self.value_formatter.get_col1_paragraph("Event date"),
                     self.value_formatter.get_timestamp_value(timestamp=misp_event.get('timestamp', None))])

        # Published
        data.append([self.value_formatter.get_col1_paragraph("Published"),
                     self.value_formatter.get_published_value(published_bool=misp_event.get('published', None),
                                                              published_timestamp=misp_event.get('publish_timestamp', None))])

        # Creator organisation
        data.append([self.value_formatter.get_col1_paragraph("Creator Org"),
                     self.value_formatter.get_creator_organisation_value(creator=misp_event.get('Orgc', None))])

        # Number of Attributes
        data.append([self.value_formatter.get_col1_paragraph("# Attributes"),
                     self.value_formatter.get_attributes_number_value(attributes=misp_event.get('Attribute', None))])

        # Tags
        curr_Tags = Tags(self.config, self.value_formatter)
        data.append([self.value_formatter.get_col1_paragraph("Tags"),
                     curr_Tags.get_tag_value(tags=misp_event.get('Tag', None))])

        flowable_table.append(create_flowable_table_from_data(data))

        # Correlation
        if is_safe_table(misp_event.get('RelatedEvent', None)) and is_in_config(self.config, 4):
            flowable_table += self.get_correlation_values(related_events=misp_event.get('RelatedEvent', None))

        # Galaxies
        if is_safe_attribute_table(misp_event, "Related Galaxies") and is_in_config(self.config, 3):
            flowable_table.append(PageBreak())
            curr_Galaxy = Galaxy(self.config, self.value_formatter)
            flowable_table += curr_Galaxy.get_galaxy_value(galaxies=misp_event.get('Galaxy', None))

        return flowable_table

    def create_reduced_flowable_table_from_event(self, misp_event):
        '''
        Returns Table presenting a MISP event
        :param misp_event: A misp event (complete or not)
        :return: a table that can be added to a pdf
        '''

        data = []
        flowable_table = []

        # Manual addition
        # UUID
        data.append([self.value_formatter.get_col1_paragraph("UUID"),
                     self.value_formatter.get_value_link_to_event(uuid=misp_event.get('uuid', None),
                                                                  text=misp_event.get('uuid', None))])

        # Info
        data.append([self.value_formatter.get_col1_paragraph("Info"),
                     self.value_formatter.get_value_link_to_event(uuid=misp_event.get('uuid', None),
                                                                  text=misp_event.get('info', None))])

        # Timestamp
        data.append([self.value_formatter.get_col1_paragraph("Event date"),
                     self.value_formatter.get_timestamp_value(timestamp=misp_event.get('timestamp', None))])

        flowable_table.append(create_flowable_table_from_data(data))

        return flowable_table

    def create_flowable_description_from_event(self, misp_event):
        '''
        Returns a Paragraph presenting a MISP event
        :param misp_event: A misp event (complete or not)
        :return: a paragraph that can be added to a pdf
        '''

        '''
        The event "{EventName}" | that occurred on {EventDate}, | had been shared by {Organisation Name} | on the {Date}.
        '''

        text = ""

        if is_safe_attribute(misp_event, 'info'):
            text += "The event '"
            text += safe_string(misp_event.info)
            text += "'"
        else:
            text += "This event"

        if is_safe_attribute(misp_event, 'timestamp'):
            text += " that occurred on "
            text += safe_string(misp_event.timestamp.strftime(EXPORT_DATE_FORMAT))
            text += ","

        text += " had been shared by "
        if is_safe_attribute(misp_event, 'Orgc') and is_safe_attribute(misp_event, 'name'):
            text += safe_string(misp_event.Orgc.name)
        else:
            text += " an unknown organisation"

        if is_safe_attribute(misp_event, 'date'):
            text += " on the "
            text += safe_string(misp_event.date)
        else:
            text += " on an unknown date"
        text += "."

        '''
        The threat level of this event is {ThreatLevel} and the analysis that was made of this event is {AnalysisLevel}.
        '''

        text += " The threat level of this event is "
        if is_safe_attribute(misp_event, 'threat_level_id') and safe_string(misp_event.threat_level_id) in threat_map:
            text += threat_map[safe_string(misp_event.threat_level_id)]
        else:
            text += " unknown"

        text += " and the analysis that was made of this event is "
        if is_safe_attribute(misp_event, 'analysis') and safe_string(misp_event.analysis) in analysis_map:
            text += analysis_map[safe_string(misp_event.analysis)]
        else:
            text += " undefined"
        text += "."

        '''
        The event is currently {Published} and has associated attributes {Attribute Number}.
        '''

        text += " The event is currently "
        if is_safe_attribute(misp_event, 'published') and misp_event.published:
            text += " published"
            if is_safe_attribute(misp_event, 'publish_timestamp'):
                text += " since " + misp_event.publish_timestamp.strftime(EXPORT_DATE_FORMAT)
        else:
            text += " private"

        # Number of Attributes
        text += ", has "
        if is_safe_attribute_table(misp_event, 'Attribute'):
            text += safe_string(len(misp_event.Attribute))
        else:
            text += " 0"

        text += " associated attributes"

        # Number of Objects
        text += " and has "
        if is_safe_attribute_table(misp_event, 'Object'):
            text += safe_string(len(misp_event.Object))
        else:
            text += " 0"

        text += " associated objects."

        curr_attributes = Attributes(self.config, self.value_formatter)
        tmp_text = curr_attributes.get_external_analysis(misp_event)

        if tmp_text != "":
            text += "<br/>"
            text += tmp_text
            text += "<br/>"

        '''
        For more information on the event, please consult the rest of the document
        '''
        text += "<br/>For more information on the event, please consult following information."

        description_style = ParagraphStyle(name='Description', parent=self.value_formatter.col2_style,
                                           alignment=TA_JUSTIFY)

        return Paragraph(text, description_style)

    def get_correlation_values(self, related_events=None):
        '''
        Returns a flowable paragraph to add to the pdf given the misp_event correlated events
        :return: a Paragraph to add in the pdf, regarding the values of "RelatedEvent"
        '''
        flowable_table = []

        flowable_table.append(PageBreak())
        flowable_table.append(Paragraph("Related Events", self.sample_style_sheet['Heading3']))

        if is_safe_table(related_events):
            for i, evt in enumerate(related_events):
                flowable_table.append(Indenter(left=INDENT_SIZE_HEADING))
                flowable_table.append(
                    Paragraph("Related Event #" + str(i + OFFSET), self.sample_style_sheet['Heading4']))
                flowable_table.append(Indenter(left=-INDENT_SIZE_HEADING))

                flowable_table += self.create_reduced_flowable_table_from_event(evt)
                i += 1
        else:
            return flowable_table.append(self.value_formatter.get_unoverflowable_paragraph(DEFAULT_VALUE))

        return flowable_table


class Attributes():

    # ----------------------------------------------------------------------
    def __init__(self, config, value_formatter):
        self.config = config
        self.value_formatter = value_formatter
        self.sample_style_sheet = getSampleStyleSheet()

    # ----------------------------------------------------------------------

    def create_flowable_table_from_attributes(self, misp_event):
        '''
        Returns a list of flowables representing the list of attributes of a misp event.
        The list is composed alternatively of headers and tables, to add to the pdf
        :param misp_event: A misp event
        :return: a table of flowables
        '''
        flowable_table = []
        i = 0

        if is_safe_attribute_table(misp_event, "Attribute"):
            # There is some attributes for this object
            for item in getattr(misp_event, "Attribute"):
                # you can use a spacer instead of title to separate paragraph: flowable_table.append(Spacer(1, 5 * mm))
                flowable_table.append(Indenter(left=INDENT_SIZE_HEADING))
                flowable_table.append(Paragraph("Attribute #" + str(i + OFFSET), self.sample_style_sheet['Heading4']))
                flowable_table.append(Indenter(left=-INDENT_SIZE_HEADING))

                flowable_table += self.create_flowable_table_from_one_attribute(item)
                i += 1
        else:
            # No attributes for this object
            flowable_table.append(Paragraph("No attributes", self.sample_style_sheet['Heading4']))

        return flowable_table

    def create_flowable_table_from_one_attribute(self, misp_attribute):
        '''
        Returns a table (flowalbe) representing the attribute
        :param misp_attribute: A misp attribute
        :return: a table representing this misp's attribute's attributes, to add to the pdf as a flowable
        '''

        data = []
        flowable_table = []

        # To reduce code size, and automate it a bit, triplet (Displayed Name, object_attribute_name,
        # to_display_if_not_present) are store in the following list
        list_attr_automated = [["UUID", 'uuid', "None"],
                               ["Category", 'category', "None"],
                               ["Comment", 'comment', "None"],
                               ["Type", 'type', "None"],
                               ["Value", 'value', "None"]]

        # Handle the special case of links
        STANDARD_TYPE = True
        if is_safe_value(misp_attribute.get('type', None)) and (misp_attribute.type in [LINK_TYPE, URL_TYPE]):
            # Special case for links
            STANDARD_TYPE = False

        # Automated adding of standard (python) attributes of the misp event
        for item in list_attr_automated:
            if is_safe_attribute(misp_attribute, item[1]) and (STANDARD_TYPE or item[1] != 'value'):
                # The attribute exists, we fetch it and create the row
                data.append([self.value_formatter.get_col1_paragraph(item[0]),
                             self.value_formatter.get_unoverflowable_paragraph(getattr(misp_attribute, item[1]))])

            # The attribute does not exist, you may want to print a default text on the row. Then use as a else case :
            # data.append([Paragraph(item[0], col1_style), Paragraph(item[2], col2_style)])

        # Handle Special case for links (Value) - There were not written in the previous loop
        if not STANDARD_TYPE and is_safe_value(misp_attribute.get('value', None)):
            data.append([self.value_formatter.get_col1_paragraph("Value"),
                         self.value_formatter.get_good_or_bad_link(value=misp_attribute.get('value', None),
                                                                   type=misp_attribute.get('type', None))])

        # Handle pictures
        if is_safe_value(misp_attribute.get('data', None)) and misp_attribute.type == IMAGE_TYPE:
            data.append([self.value_formatter.get_col1_paragraph("Data"),
                         self.value_formatter.get_image_value(misp_attribute.get('data', None))])

        # Tags
        curr_Tags = Tags(self.config, self.value_formatter)

        if is_safe_table(misp_attribute.get('Tag', None)):
            data.append(
                [self.value_formatter.get_col1_paragraph("Tags"),
                 curr_Tags.get_tag_value(tags=misp_attribute.get('Tag', None))])

        # Sighting
        curr_Sighting = Sightings(self.config, self.value_formatter)

        if is_safe_table(misp_attribute.get('Sighting', None)):
            data.append([self.value_formatter.get_col1_paragraph("Sighting"),
                         curr_Sighting.create_flowable_paragraph_from_sightings(sightings=misp_attribute.get('Sighting', None))])

        flowable_table.append(create_flowable_table_from_data(data))

        # Galaxies
        if is_safe_attribute_table(misp_attribute, "Galaxy") and is_in_config(self.config, 3):
            curr_Galaxy = Galaxy(self.config, self.value_formatter)
            flowable_table.append(Indenter(left=INDENT_SIZE))
            flowable_table += curr_Galaxy.get_galaxy_value(misp_attribute.get('Galaxy', None))
            flowable_table.append(Indenter(left=-INDENT_SIZE))

        return flowable_table

    def get_external_analysis(self, misp_event):
        '''
        Returns a string representing the list of external analysis comments of a misp event.
        :param misp_event: A misp event
        :return: a table of flowables
        '''
        text = ""

        if is_safe_attribute_table(misp_event, "Attribute"):
            # There is some attributes for this object
            for attribute in getattr(misp_event, "Attribute"):
                # If the current event is an external analysis and a comment
                if (is_safe_attribute(attribute, "value")
                        and is_safe_attribute(attribute, "category")
                        and is_safe_attribute(attribute, "type")
                        and getattr(attribute, "category") == "External analysis"
                        and getattr(attribute, "type") == "comment"):
                    # We add it to the description
                    text += "<br/>" + EXTERNAL_ANALYSIS_PREFIX + safe_string(getattr(attribute, "value"))

        return text


class Tags():

    # ----------------------------------------------------------------------
    def __init__(self, config, value_formatter):
        self.config = config
        self.value_formatter = value_formatter

    # ----------------------------------------------------------------------

    def get_tag_value(self, tags=None):
        '''
        Returns a flowable paragraph to add to the pdf given the misp_event tags
        :return: a Paragraph to add in the pdf, regarding the values of "tags"
        '''
        if is_safe_table(tags):
            table_event_tags = self.create_flowable_table_from_tags(tags=tags)
            return table_event_tags
        return self.value_formatter.get_unoverflowable_paragraph(DEFAULT_VALUE)

    def create_flowable_table_from_tags(self, tags=None):
        '''
        Returns a Table (flowable) to add to a pdf, representing the list of tags of an event or a misp event
        :param misp_event: A misp event
        :return: a table of flowable to add to the pdf
        '''

        flowable_table = []
        i = 0

        if is_safe_table(tags):
            # There is some tags for this object
            for curr_tag in tags:
                flowable_table.append(create_flowable_tag(curr_tag))
                i += 1
            answer_tags = self.create_tags_table_from_data(flowable_table)
        else:
            # No tags for this object
            answer_tags = [self.value_formatter.get_unoverflowable_paragraph("No tags")]

        return answer_tags

    def create_tags_table_from_data(self, data):
        '''
        Given a list of flowables tags (2D/list of list), creates a Table with styles adapted to tags.
        :param data: list of list of tags (flowables)
        :return: a Table - with styles - to add to another table
        '''

        # Create the table
        curr_table = Table(data, COL_WIDTHS, rowHeights=ROW_HEIGHT_FOR_TAGS)

        # Create styles and set parameters
        general_style = general_style_generator()

        # Make the table nicer
        curr_table.setStyle(TableStyle(general_style))

        return curr_table


class Sightings():

    # ----------------------------------------------------------------------
    def __init__(self, config, value_formatter):
        self.config = config
        self.value_formatter = value_formatter

    # ----------------------------------------------------------------------

    def create_flowable_paragraph_from_sightings(self, sightings=None):
        '''
        Returns a Table (flowable) to add to a pdf, representing the list of sightings of an event or a misp event
        :param misp_event: A misp event
        :return: a table of flowable to add to the pdf
        '''

        i = 0

        # No tags for this object
        answer = "No sighting"

        list_sighting = [0, 0, 0]
        if is_safe_table(sightings):
            # There is some tags for this object
            for curr_item in sightings:
                # TODO : When Sightings will be object : if is_safe_attribute(item, "type"):
                if is_safe_dict_attribute(curr_item, "type"):
                    # Store the likes/dislikes depending on their types
                    list_sighting[int(curr_item["type"])] += 1
                i += 1

            # Create the sighting text
            sight_text = f"<font color ={POSITIVE_SIGHT_COLOR}> Positive: {list_sighting[0]}</font>"
            sight_text += f" / <font color ={NEGATIVE_SIGHT_COLOR}> Negative: {list_sighting[1]}</font>"
            sight_text += f" / <font color ={MISC_SIGHT_COLOR}> Misc.: {list_sighting[2]}</font>"

            answer = sight_text

        return self.value_formatter.get_unoverflowable_paragraph(answer, do_escape_string=False)


class Object():

    # ----------------------------------------------------------------------
    def __init__(self, config, value_formatter):
        self.config = config
        self.value_formatter = value_formatter
        self.sample_style_sheet = getSampleStyleSheet()

    # ----------------------------------------------------------------------

    def create_flowable_table_from_objects(self, objects=None):
        '''
        Returns a list of flowables representing the list of objects of a misp event.
        The list is composed of a serie of
        [ header object, table of object information, [ header of attribute, table of attribute]*] to add to the pdf
        :param misp_event: A misp event
        :return: a table of flowables
        '''

        flowable_table = []
        i = 0
        if is_safe_table(objects):

            # There is a list of objects
            for item in objects:
                # you can use a spacer instead of title to separate paragraph: flowable_table.append(Spacer(1, 5 * mm))
                flowable_table.append(Indenter(left=INDENT_SIZE_HEADING))
                flowable_table.append(Paragraph("Object #" + str(i + OFFSET), self.sample_style_sheet['Heading3']))
                flowable_table.append(Indenter(left=-INDENT_SIZE_HEADING))
                flowable_table += self.create_flowable_table_from_one_object(item, self.config)
                i += 1
        else:
            # No object found
            flowable_table.append(Paragraph("No object", self.sample_style_sheet['Heading3']))

        return flowable_table

    def create_flowable_table_from_one_object(self, misp_object, config=None):
        '''
        Returns a table (flowable) representing the object
        :param misp_attribute: A misp object
        :return: a table representing this misp's object's attributes, to add to the pdf as a flowable
        '''
        data = []

        # To reduce code size, and automate it a bit, triplet (Displayed Name, object_attribute_name,
        # to_display_if_not_present) are store in the following list
        list_attr_automated = [["UUID", 'uuid', "None"],
                               ["Description", 'description', "None"],
                               ["Meta Category", 'meta-category', "None"],
                               ["Object Name", 'name', "None"],
                               ["Comment", 'comment', "None"],
                               ["Type", 'type', "None"]]

        # Automated adding of standard (python) attributes of the misp object
        for item in list_attr_automated:
            if is_safe_attribute(misp_object, item[1]):
                # The attribute exists, we fetch it and create the row
                data.append([self.value_formatter.get_col1_paragraph(item[0]),
                             self.value_formatter.get_unoverflowable_paragraph(getattr(misp_object, item[1]))])

            # The attribute does not exist, you may want to print a default text on the row. Then use as a else case :
            # data.append([Paragraph(item[0], col1_style), Paragraph(item[2], col2_style)])

        # Timestamp
        data.append([self.value_formatter.get_col1_paragraph("Object date"),
                     self.value_formatter.get_timestamp_value(misp_object.get('timestamp', None))])

        # Transform list of value in a table
        data = [create_flowable_table_from_data(data)]

        # Handle all the attributes
        if is_safe_value(misp_object.get("Attribute", None)):
            curr_attributes = Attributes(self.config, self.value_formatter)
            data.append(Indenter(left=INDENT_SIZE))
            data += curr_attributes.create_flowable_table_from_attributes(misp_object)
            data.append(Indenter(left=-INDENT_SIZE))

        # Add a page break at the end of an object
        data.append(PageBreak())

        return data


class Galaxy():

    # ----------------------------------------------------------------------
    def __init__(self, config, value_formatter):
        self.config = config
        self.value_formatter = value_formatter
        self.sample_style_sheet = getSampleStyleSheet()

    # ----------------------------------------------------------------------

    def get_galaxy_value(self, galaxies=None):
        '''
        Returns a flowable paragraph to add to the pdf given the misp_event galaxies
        :return: a Flowable to add in the pdf, regarding the values of "galaxies"
        '''

        flowable_table = []

        # Galaxies
        # item = ["Related Galaxies", 'Galaxy', "None"]
        if is_safe_table(galaxies) and is_in_config(self.config, 3):
            galaxy_title = Paragraph(safe_string("Related Galaxies"), self.sample_style_sheet['Heading5'])
            flowable_table.append(Indenter(left=INDENT_SIZE_HEADING))
            flowable_table.append(galaxy_title)
            flowable_table.append(Indenter(left=-INDENT_SIZE_HEADING))
            flowable_table += self.create_flowable_table_from_galaxies(galaxies=galaxies)
        else:
            flowable_table.append(self.value_formatter.get_unoverflowable_paragraph(DEFAULT_VALUE))

        return flowable_table

    def create_flowable_table_from_galaxies(self, galaxies=None):
        '''
        Returns a Table (flowable) to add to a pdf, representing the list of galaxies of an event or a misp event
        :param misp_event: A misp event
        :return: a table of flowables to add to the pdf
        '''
        flowable_table = []
        scheme_alternation = []
        curr_color = 0
        i = 0
        small_title_style = ParagraphStyle(name='Column_1', parent=self.sample_style_sheet['Heading6'],
                                           fontName=FIRST_COL_FONT, alignment=TA_LEFT)

        if is_safe_table(galaxies):
            # There is some galaxies for this object

            for curr_galaxy in galaxies:
                # For each galaxy of the misp object

                txt_title = "Galaxy #" + str(i + OFFSET) + " - " + safe_string(curr_galaxy["name"])
                galaxy_title = Paragraph(txt_title, small_title_style)
                flowable_table.append(Indenter(left=INDENT_SIZE_HEADING))
                flowable_table.append(galaxy_title)
                flowable_table.append(Indenter(left=-INDENT_SIZE_HEADING))
                i += 1

                # Add metadata about the Galaxy
                galaxy_metadata, nb_added_item = self.create_flowable_table_from_one_galaxy(curr_galaxy)
                flowable_table += galaxy_metadata

                # Construct the line color scheme and line scheme
                scheme_alternation += [curr_color] * nb_added_item

                # Add metadata about clusters
                curr_cluster = Galaxy_cluster(self.config, self.value_formatter)
                clusters_metadata = curr_cluster.create_flowable_table_from_galaxy_clusters(curr_galaxy)
                flowable_table += clusters_metadata

        else:
            # No galaxies for this object
            answer_tags = [self.value_formatter.get_unoverflowable_paragraph("No galaxies")]
            flowable_table.append(create_flowable_table_from_data(answer_tags))

        return flowable_table

    def create_flowable_table_from_one_galaxy(self, misp_galaxy):
        '''
        Returns a table (flowable) representing the galaxy
        :param misp_galaxy: A misp galaxy
        :return: a table representing this misp's galaxy's attributes, to add to the pdf as a flowable
        '''
        data = []
        nb_added_item = 0

        # Name
        if is_safe_value(misp_galaxy.get('name', None)):
            data.append([self.value_formatter.get_col1_paragraph("Name", do_small=DO_SMALL_GALAXIES),
                         self.value_formatter.get_galaxy_name_value(misp_galaxy)])
            nb_added_item += 1

        # Description
        if is_safe_value(misp_galaxy.get('description', None)):
            data.append([self.value_formatter.get_col1_paragraph("Description", do_small=DO_SMALL_GALAXIES),
                         self.value_formatter.get_unoverflowable_paragraph(misp_galaxy.get('description', None), do_small=DO_SMALL_GALAXIES)])
            nb_added_item += 1

        flowable_table = []
        flowable_table.append(create_flowable_table_from_data(data))

        return flowable_table, nb_added_item


class Galaxy_cluster():

    # ----------------------------------------------------------------------
    def __init__(self, config, value_formatter):
        self.config = config
        self.value_formatter = value_formatter
        self.sample_style_sheet = getSampleStyleSheet()

    # ----------------------------------------------------------------------
    def create_flowable_table_from_galaxy_clusters(self, misp_galaxy):
        '''
        Returns a Table (flowable) to add to a pdf, representing the list of galaxy clusters of a galaxy
        :param misp_event: A misp event
        :return: a table of flowables to add to the pdf
        '''

        data = []
        tmp_title = "Cluster #"

        if is_safe_value(misp_galaxy.get("GalaxyCluster", None)):
            # There is some clusters for this object
            for i, curr_cluster in enumerate(misp_galaxy.get("GalaxyCluster", None)):
                # If title is needed :
                # galaxy_title = [Paragraph("Cluster #" + str(i), self.sample_style_sheet['Heading6'])]
                # data.append(galaxy_title)

                tmp_title = "Cluster #" + str(i + OFFSET)

                # For each cluster
                tmp_data = self.create_flowable_table_from_one_galaxy_cluster(curr_cluster)
                tmp_flowable_table = []
                tmp_flowable_table.append(create_flowable_table_from_data(tmp_data, col_w=SECOND_LEVEL_GALAXY_WIDTHS,
                                                                          color_alternation=CLUSTER_COLORS,
                                                                          line_alternation=[], galaxy_colors=True))
                data.append([self.value_formatter.get_col1_paragraph(tmp_title, do_small=DO_SMALL_GALAXIES),
                             tmp_flowable_table])  # Cluster #X - 3 lines

        else:
            # No galaxies for this object
            data = [self.value_formatter.get_unoverflowable_paragraph("No galaxy cluster", do_small=DO_SMALL_GALAXIES)]

        flowable_table = []
        flowable_table.append(
            create_flowable_table_from_data(data, col_w=FIRST_LEVEL_GALAXY_WIDTHS, color_alternation=CLUSTER_COLORS,
                                            galaxy_colors=True))

        return flowable_table

    def create_flowable_table_from_one_galaxy_cluster(self, misp_cluster):
        '''
        Returns a table (flowable) representing a galaxy cluster
        :param misp_attribute: A misp galaxy
        :return: a table representing this misp's galaxy's cluster attributes, to add to the pdf as a flowable
        '''
        data = []

        # Name
        data.append([self.value_formatter.get_col1_paragraph("Name", do_small=True),
                     self.value_formatter.get_galaxy_cluster_name_value(misp_cluster, do_small=True)])

        if misp_cluster['value'] != misp_cluster['description']:  # Prevent name that are same as description
            # Description
            data.append([self.value_formatter.get_col1_paragraph("Description", do_small=True),
                         self.value_formatter.get_unoverflowable_paragraph(misp_cluster.get('description', None), do_small=True)])

        # Refs ?
        # item = ["Description", 'description', "None"]
        # data.append([self.value_formatter.get_col1_paragraph(item[0]),
        # self.value_formatter.get_unoverflowable_paragraph(misp_cluster[item[1]])])

        return data


########################################################################
# Handling static parts drawn on the upper layer


class Statics_Drawings():

    # ----------------------------------------------------------------------
    def __init__(self, config, misp_event):
        self.config = config
        self.misp_event = misp_event

    # ----------------------------------------------------------------------

    def set_template(self, canvas, doc):
        self.add_page_number(canvas, doc)
        self.add_metadata(canvas, doc)
        # TODO : add_header()
        # TODO : add_footer()

    def add_metadata(self, canvas, doc):
        '''
        Allow to add metadata to the pdf. Would need deeper digging to change other metadata.
        :param canvas: / Automatically filled during pdf compilation
        :param doc: / Automatically filled during pdf compilation
        :param misp_event: To send trough "partial", to get information to complete metadaa
        :return: / Automatically filled during pdf compilation
        '''

        if is_safe_attribute(self.misp_event, 'info'):
            canvas.setTitle(self.misp_event.info)
            canvas.setSubject(self.misp_event.info)

        if is_safe_attribute(self.misp_event, 'Orgc'):
            if is_safe_attribute(self.misp_event.Orgc, 'name'):
                canvas.setAuthor(self.misp_event.Orgc.name)

                if is_in_config(self.config, 1):
                    canvas.setCreator(self.config[moduleconfig[1]])
                else:
                    canvas.setCreator(self.misp_event.Orgc.name)

        if is_safe_attribute(self.misp_event, 'uuid'):
            canvas.setKeywords(self.misp_event.uuid)

    def add_page_number(self, canvas, doc):
        '''
        Draw the page number on each page
        :param canvas: / Automatically filled during pdf compilation
        :param doc: / Automatically filled during pdf compilation
        :return: / Automatically filled during pdf compilation
        '''
        canvas.saveState()
        canvas.setFont('Times-Roman', 10)
        page_number_text = "%d" % (doc.page)

        curr_spacing = 4 * mm  # 0.75 * inch

        canvas.drawCentredString(
            curr_spacing,
            curr_spacing,
            page_number_text
        )

        canvas.restoreState()


########################################################################
# Main part of the script, handling the global formatting of the pdf

def collect_parts(misp_event, config=None):
    '''
    Main part of the PDF creation, it creates a ready-to-compile-as-pdf list of flowables from a MISP Event, calling subfunctions to handle the printing of each element
    :param misp_event: a misp event
    :return: a list of flowables to compile as pdf
    '''
    # List of elements/content we want to add
    flowables = []

    # Get the list of available styles
    sample_style_sheet = getSampleStyleSheet()
    col1_style, col2_style = get_table_styles()
    col1_small_style, col2_small_style = get_clusters_table_styles()
    curr_val_f = Value_Formatter(config, col1_style, col2_style, col1_small_style, col2_small_style)

    # Create stuff
    title_style = ParagraphStyle(name='Column_1', parent=sample_style_sheet['Heading1'],
                                 fontName=FIRST_COL_FONT, alignment=TA_CENTER)
    title = curr_val_f.get_value_link_to_event(uuid=misp_event.get('uuid', None),
                                               text=misp_event.get('info', None),
                                               curr_style=title_style, color=False)
    # Add all parts to final PDF
    flowables.append(title)

    # Creation of handling objects
    curr_event = Event_Metadata(config, curr_val_f)
    curr_attr = Attributes(config, curr_val_f)
    curr_object = Object(config, curr_val_f)

    if is_in_config(config, 2):  # If description is activated
        description = Paragraph("Description", sample_style_sheet['Heading2'])
        description_text = curr_event.create_flowable_description_from_event(misp_event)
        flowables.append(description)
        flowables.append(description_text)

    subtitle = Paragraph("General information", sample_style_sheet['Heading2'])
    table_general_metainformation = curr_event.create_flowable_table_from_event(misp_event)
    flowables.append(subtitle)
    flowables += table_general_metainformation

    if is_safe_attribute_table(misp_event, "Attribute"):
        flowables.append(PageBreak())

    event_attributes_title = Paragraph("Attributes", sample_style_sheet['Heading2'])
    table_direct_attributes = curr_attr.create_flowable_table_from_attributes(misp_event)
    flowables.append(event_attributes_title)
    flowables += table_direct_attributes

    if is_safe_attribute_table(misp_event, "Object"):
        flowables.append(PageBreak())

    event_objects_title = Paragraph("Objects", sample_style_sheet['Heading2'])
    table_objects = curr_object.create_flowable_table_from_objects(objects=misp_event.get("Object", None))
    flowables.append(event_objects_title)
    flowables += table_objects

    # If you want to output the full json (as debug), just add next line and add it to flowables
    # paragraph_2 = Paragraph(str(misp_event.to_json()), sample_style_sheet['Code'])

    return flowables


def export_flowables_to_pdf(document, misp_event, flowables, config):
    '''
    Export function : creates a pdf from a list of flowables, adding page numbers, etc.
    :param document: A document template
    :param pdf_buffer: / not used
    :param flowables: list of flowables to compile as pdf
    :return:
    '''

    static_drawer = Statics_Drawings(config, misp_event)

    document.build(
        flowables,
        # Partial used to set the metadata
        onFirstPage=static_drawer.set_template,  # Pagination for first page
        onLaterPages=static_drawer.set_template,  # Pagination for all other page
    )
    # Old way : onLaterPages=partial(static_drawer.set_template, misp_event=misp_event),  # Pagination for all other page


########################################################################
# "EXTERNAL" exposed METHODS. Meant to be used outside of this class.

def convert_event_in_pdf_buffer(misp_event, config=None):
    '''
    Externally callable function that create a full pdf from a Misp Event
    :param misp_event: a misp event
    :return: a pdf buffer (BytesIO) that contains the pdf
    '''
    # Create a document buffer
    pdf_buffer = BytesIO()

    if is_in_config(config, 5):  # We want internationalization
        logger.info("Internationalization of fonts during pdf export activated. CJK-fonts supported.")
        internationalize_font(config)

    # DEBUG / TO DELETE : curr_document = SimpleDocTemplate('myfile.pdf')
    curr_document = SimpleDocTemplate(pdf_buffer,
                                      pagesize=PAGESIZE,
                                      topMargin=BASE_MARGIN,
                                      leftMargin=BASE_MARGIN,
                                      rightMargin=BASE_MARGIN,
                                      bottomMargin=BASE_MARGIN)

    # Collect already accessible event's parts to be shown
    flowables = collect_parts(misp_event, config)

    # Export
    export_flowables_to_pdf(curr_document, misp_event, flowables, config)
    pdf_value = pdf_buffer.getvalue()

    # Not sure what to give back ? Buffer ? Buffer.value() ? Base64(buffer.value()) ? ... So far only buffer.value()
    pdf_buffer.close()

    return pdf_value


def get_values_from_buffer(pdf_buffer):
    return pdf_buffer.value()


def get_base64_from_buffer(pdf_buffer):
    return base64.b64encode(pdf_buffer.value())


def get_base64_from_value(pdf_value):
    return base64.b64encode(pdf_value)


def register_to_file(pdf_buffer, file_name):
    # Used for testing purposes only
    pdf_buffer.seek(0)

    with open(file_name, 'wb') as f:
        f.write(pdf_buffer.read())


def register_value_to_file(pdf_value, file_name):
    with open(file_name, 'wb') as f:
        f.write(pdf_value)



__version__ = '2.4.162.1'
import logging
import sys
import warnings

logger = logging.getLogger(__name__)


def warning_2022():
    if sys.version_info < (3, 8):
        warnings.warn("""
As our baseline system is the latest Ubuntu LTS, and Ubuntu LTS 20.04 has Python 3.8 available,
we will officially deprecate python versions below 3.8 on January 1st 2022.
**Please update your codebase.**""", DeprecationWarning, stacklevel=3)


everything_broken = '''Unknown error: the response is not in JSON.
Something is broken server-side, please send us everything that follows (careful with the auth key):
Request headers:
{}
Request body:
{}
Response (if any):
{}'''


try:
    warning_2022()
    from .exceptions import PyMISPError, NewEventError, NewAttributeError, MissingDependency, NoURL, NoKey, InvalidMISPObject, UnknownMISPObjectTemplate, PyMISPInvalidFormat, MISPServerError, PyMISPNotImplementedYet, PyMISPUnexpectedResponse, PyMISPEmptyResponse  # noqa
    from .abstract import AbstractMISP, MISPEncode, pymisp_json_default, MISPTag, Distribution, ThreatLevel, Analysis  # noqa
    from .mispevent import (MISPEvent, MISPAttribute, MISPObjectReference, MISPObjectAttribute, MISPObject, MISPUser, # noqa
                            MISPOrganisation, MISPSighting, MISPLog, MISPShadowAttribute, MISPWarninglist, MISPTaxonomy,
                            MISPNoticelist, MISPObjectTemplate, MISPSharingGroup, MISPRole, MISPServer, MISPFeed,
                            MISPEventDelegation, MISPUserSetting, MISPInbox, MISPEventBlocklist, MISPOrganisationBlocklist,
                            MISPEventReport, MISPGalaxyCluster, MISPGalaxyClusterElement, MISPGalaxyClusterRelation,
                            MISPCorrelationExclusion, MISPGalaxy)
    from .tools import AbstractMISPObjectGenerator  # noqa
    from .tools import Neo4j  # noqa
    from .tools import stix  # noqa
    from .tools import openioc  # noqa
    from .tools import ext_lookups  # noqa
    from .tools import update_objects  # noqa

    from .api import PyMISP, register_user  # noqa
    from .api import PyMISP as ExpandedPyMISP  # noqa
    from .tools import load_warninglists  # noqa
    # Let's not bother with old python
    try:
        from .tools import reportlab_generator  # noqa
    except ImportError:
        # FIXME: The import should not raise an exception if reportlab isn't installed
        pass
    except NameError:
        # FIXME: The import should not raise an exception if reportlab isn't installed
        pass
    logger.debug('pymisp loaded properly')
except ImportError as e:
    logger.warning('Unable to load pymisp properly: {}'.format(e))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import re
import logging
import ipaddress
import email.utils
from email import policy, message_from_bytes
from email.message import EmailMessage
from io import BytesIO
from pathlib import Path
from typing import Union, List, Tuple, Dict, cast, Any

from extract_msg import openMsg
from extract_msg.message import Message as MsgObj
from RTFDE.exceptions import MalformedEncapsulatedRtf, NotEncapsulatedRtf  # type: ignore
from RTFDE.deencapsulate import DeEncapsulator  # type: ignore
from oletools.common.codepages import codepage2codec  # type: ignore

from ..exceptions import InvalidMISPObject, PyMISPNotImplementedYet, MISPObjectException, NewAttributeError
from .abstractgenerator import AbstractMISPObjectGenerator

logger = logging.getLogger('pymisp')


class MISPMsgConverstionError(MISPObjectException):
    pass


class EMailObject(AbstractMISPObjectGenerator):
    def __init__(self, filepath: Union[Path, str] = None, pseudofile: BytesIO = None,
                 attach_original_email: bool = True, **kwargs):
        super().__init__('email', **kwargs)

        self.attach_original_email = attach_original_email
        self.encapsulated_body: Union[str, None] = None
        self.eml_from_msg: Union[bool, None] = None
        self.raw_emails: Dict[str, Union[BytesIO, None]] = {'msg': None,
                                                            'eml': None}

        self.__pseudofile = self.create_pseudofile(filepath, pseudofile)
        self.email = self.parse_email()
        self.generate_attributes()

    def parse_email(self) -> EmailMessage:
        """Convert email into EmailMessage."""
        content_in_bytes = self.__pseudofile.getvalue().strip()
        eml = message_from_bytes(content_in_bytes,
                                 _class=EmailMessage,
                                 policy=policy.default)
        eml = cast(EmailMessage, eml)  # Only needed to quiet mypy
        if len(eml) != 0:
            self.raw_emails['eml'] = self.__pseudofile
            return eml
        else:
            logger.debug("Email not in standard .eml format. Attempting to decode email from other formats.")
        try:  # Check for .msg formatted emails.
            # Msg files have the same header signature as the CFB format
            if content_in_bytes[:8] == b"\xd0\xcf\x11\xe0\xa1\xb1\x1a\xe1":
                message = self._msg_to_eml(content_in_bytes)
                if len(message) != 0:
                    self.eml_from_msg = True
                    self.raw_emails['msg'] = self.__pseudofile
                    self.raw_emails['msg'] = BytesIO(message.as_bytes())
                    return message
        except ValueError as _e:  # Exception
            logger.debug("Email not in .msg format or is a corrupted .msg. Attempting to decode email from other formats.")
            logger.debug("Error: {} ".format(_e))
        try:
            if content_in_bytes[:3] == b'\xef\xbb\xbf':  # utf-8-sig byte-order mark (BOM)
                eml_bytes = content_in_bytes.decode("utf_8_sig").encode("utf-8")
                eml = email.message_from_bytes(eml_bytes,
                                               policy=policy.default)
                eml = cast(EmailMessage, eml)  # Only needed to quiet mypy
                if len(eml) != 0:
                    self.raw_emails['eml'] = BytesIO(eml_bytes)
                    return eml
        except UnicodeDecodeError:
            pass
        raise PyMISPNotImplementedYet("EmailObject does not know how to decode data passed to it. Object may not be an email. If this is an email please submit it as an issue to PyMISP so we can add support.")

    @staticmethod
    def create_pseudofile(filepath: Union[Path, str] = None,
                          pseudofile: BytesIO = None) -> BytesIO:
        """Creates a pseudofile using directly passed data or data loaded from file path.
        """
        if filepath:
            with open(filepath, 'rb') as f:
                return BytesIO(f.read())
        elif pseudofile and isinstance(pseudofile, BytesIO):
            return pseudofile
        else:
            raise InvalidMISPObject('File buffer (BytesIO) or a path is required.')

    def _msg_to_eml(self, msg_bytes: bytes) -> EmailMessage:
        """Converts a msg into an eml."""
        msg_obj = openMsg(msg_bytes)
        # msg obj stores the original raw header here
        message, body, attachments = self._extract_msg_objects(msg_obj)
        eml = self._build_eml(message, body, attachments)
        return eml

    def _extract_msg_objects(self, msg_obj: MsgObj) -> Tuple[EmailMessage, Dict, List[Any]]:
        """Extracts email objects needed to construct an eml from a msg."""
        message: EmailMessage = email.message_from_string(msg_obj.header.as_string(), policy=policy.default)  # type: ignore
        body = {}
        if msg_obj.body is not None:
            body['text'] = {"obj": msg_obj.body,
                            "subtype": 'plain',
                            "charset": "utf-8",
                            "cte": "base64"}
        if msg_obj.htmlBody is not None:
            try:
                _html_encoding_raw = msg_obj.mainProperties['3FDE0003'].value
                _html_encoding = codepage2codec(_html_encoding_raw)
            except KeyError:
                _html_encoding = msg_obj.stringEncoding
            body['html'] = {'obj': msg_obj.htmlBody.decode(),
                            "subtype": 'html',
                            "charset": _html_encoding,
                            "cte": "base64"}
        if msg_obj.rtfBody is not None:
            body['rtf'] = {"obj": msg_obj.rtfBody.decode(),
                           "subtype": 'rtf',
                           "charset": 'ascii',
                           "cte": "base64"}
            try:
                rtf_obj = DeEncapsulator(msg_obj.rtfBody)
                rtf_obj.deencapsulate()
                if (rtf_obj.content_type == "html") and (msg_obj.htmlBody is None):
                    self.encapsulated_body = 'text/html'
                    body['html'] = {"obj": rtf_obj.html,
                                    "subtype": 'html',
                                    "charset": rtf_obj.text_codec,
                                    "cte": "base64"}
                elif (rtf_obj.content_type == "text") and (msg_obj.body is None):
                    self.encapsulated_body = 'text/plain'
                    body['text'] = {"obj": rtf_obj.plain_text,
                                    "subtype": 'plain',
                                    "charset": rtf_obj.text_codec}
            except NotEncapsulatedRtf:
                logger.debug("RTF body in Msg object is not encapsualted.")
            except MalformedEncapsulatedRtf:
                logger.info("RTF body in Msg object contains encapsulated content, but it is malformed and can't be converted.")
        attachments = msg_obj.attachments
        return message, body, attachments

    def _build_eml(self, message: EmailMessage, body: dict, attachments: list) -> EmailMessage:
        """Constructs an eml file from objects extracted from a msg."""
        # Order the body objects by increasing complexity and toss any missing objects
        body_objects: List[dict] = [body.get('text', {}),
                                    body.get('html', {}),
                                    body.get('rtf', {})]
        body_objects = [i for i in body_objects if i != {}]
        # If this a non-multipart email then we only need to attach the payload
        if message.get_content_maintype() != 'multipart':
            for _body in body_objects:
                if "text/{0}".format(_body['subtype']) == message.get_content_type():
                    message.set_content(**_body)
                    return message
            raise MISPMsgConverstionError("Unable to find appropriate eml payload in message body.")
        # If multipart we are going to have to set the content type to null and build it back up.
        _orig_boundry = message.get_boundary()
        message.clear_content()
        # See if we are dealing with `related` inline content
        related_content = {}
        if isinstance(body.get('html', None), dict):
            _html = body.get('html', {}).get('obj')
            for attch in attachments:
                if _html.find("cid:{0}".format(attch.cid)) != -1:
                    _content_type = attch._getStringStream('__substg1.0_370E')
                    maintype, subtype = _content_type.split("/", 1)
                    related_content[attch.cid] = (attch,
                                                  {'obj': attch.data,
                                                   "maintype": maintype,
                                                   "subtype": subtype,
                                                   "cid": attch.cid,
                                                   "filename": attch.longFilename})
        if len(related_content) > 0:
            if body.get('text', None) is not None:
                # Text always goes first in an alternative, but we need the related object first
                body_text = body.get('text')
                if isinstance(body_text, dict):
                    message.add_related(**body_text)
            else:
                body_html = body.get('html')
                if isinstance(body_html, dict):
                    message.add_related(**body_html)
            for mime_items in related_content.values():
                if isinstance(mime_items[1], dict):
                    message.add_related(**mime_items[1])
                cur_attach = message.get_payload()[-1]
                self._update_content_disp_properties(mime_items[0], cur_attach)
            if body.get('text', None):
                # Now add the HTML as an alternative within the related obj
                related = message.get_payload()[0]
                related.add_alternative(**body.get('html'))
        else:
            for mime_dict in body_objects:
                # If encapsulated then don't attach RTF
                if self.encapsulated_body is not None:
                    if mime_dict.get('subtype', "") == "rtf":
                        continue
                if isinstance(mime_dict, dict):
                    message.add_alternative(**mime_dict)
        for attch in attachments:  # Add attachments at the end.
            if attch.cid not in related_content.keys():
                _content_type = attch._getStringStream('__substg1.0_370E')
                maintype, subtype = _content_type.split("/", 1)
                message.add_attachment(attch.data,
                                       maintype=maintype,
                                       subtype=subtype,
                                       cid=attch.cid,
                                       filename=attch.longFilename)
                cur_attach = message.get_payload()[-1]
                self._update_content_disp_properties(attch, cur_attach)
        message.set_boundary(_orig_boundry)  # Set back original boundary
        return message

    @staticmethod
    def _update_content_disp_properties(msg_attch, eml_attch):
        """Set Content-Disposition params on binary eml objects

        You currently have to set non-filename content-disp params by hand in python.
        """
        attch_cont_disp_props = {'30070040': "creation-date",
                                 '30080040': "modification-date"}
        for num, name in attch_cont_disp_props.items():
            try:
                eml_attch.set_param(name,
                                    email.utils.format_datetime(msg_attch.props[num].value),
                                    header='Content-Disposition')
            except KeyError:
                # It's fine if they don't have those values
                pass

    @property
    def attachments(self) -> List[Tuple[str, BytesIO]]:
        to_return = []
        try:
            for attachment in self.email.iter_attachments():
                content = attachment.get_content()  # type: ignore
                if isinstance(content, str):
                    content = content.encode()
                to_return.append((attachment.get_filename(), BytesIO(content)))
        except AttributeError:
            # ignore bug in Python3.6, that cause exception for empty email body,
            # see https://stackoverflow.com/questions/56391306/attributeerror-str-object-has-no-attribute-copy-when-parsing-multipart-emai
            pass
        return to_return

    def generate_attributes(self):

        # Attach original & Converted
        if self.attach_original_email is not None:
            self.add_attribute("eml", value="Full email.eml",
                               data=self.raw_emails.get('eml'),
                               comment="Converted from MSG format" if self.eml_from_msg else None)
            if self.raw_emails.get('msg', None) is not None:
                self.add_attribute("msg", value="Full email.msg",
                                   data=self.raw_emails.get('msg'))

        message = self.email

        for _pref, body in message._find_body(message, preferencelist=['plain', 'html']):
            comment = "{0} body".format(body.get_content_type())
            if self.encapsulated_body == body.get_content_type():
                comment += " De-Encapsulated from RTF in original msg."
            self.add_attribute("email-body",
                               body.get_content(),
                               comment=comment)

        headers = ["{}: {}".format(k, v) for k, v in message.items()]
        if headers:
            self.add_attribute("header", "\n".join(headers))

        if "Date" in message and message.get('date').datetime is not None:
            self.add_attribute("send-date", message.get('date').datetime)

        if "To" in message:
            self.__add_emails("to", message["To"])
        if "Delivered-To" in message:
            self.__add_emails("to", message["Delivered-To"])

        if "From" in message:
            self.__add_emails("from", message["From"])

        if "Return-Path" in message:
            realname, address = email.utils.parseaddr(message["Return-Path"])
            self.add_attribute("return-path", address)

        if "Reply-To" in message:
            self.__add_emails("reply-to", message["reply-to"])

        if "Bcc" in message:
            self.__add_emails("bcc", message["Bcc"])

        if "Cc" in message:
            self.__add_emails("cc", message["Cc"])

        if "Subject" in message:
            self.add_attribute("subject", message["Subject"])

        if "Message-ID" in message:
            self.add_attribute("message-id", message["Message-ID"])

        if "User-Agent" in message:
            self.add_attribute("user-agent", message["User-Agent"])

        boundary = message.get_boundary()
        if boundary:
            self.add_attribute("mime-boundary", boundary)

        if "X-Mailer" in message:
            self.add_attribute("x-mailer", message["X-Mailer"])

        if "Thread-Index" in message:
            self.add_attribute("thread-index", message["Thread-Index"])

        self.__generate_received()

    def __add_emails(self, typ: str, data: str, insert_display_names: bool = True):
        addresses = []
        display_names = []

        for realname, address in email.utils.getaddresses([data]):
            if address and realname:
                addresses.append({"value": address, "comment": "{} <{}>".format(realname, address)})
            elif address:
                addresses.append({"value": address})
            else:  # parsing failed, skip
                continue

            if realname:
                display_names.append({"value": realname, "comment": "{} <{}>".format(realname, address)})

        if addresses:
            self.add_attributes(typ, *addresses)
        if insert_display_names and display_names:
            try:
                self.add_attributes("{}-display-name".format(typ), *display_names)
            except NewAttributeError:
                # email object doesn't support display name for all email addrs
                pass

    def __generate_received(self):
        """
        Extract IP addresses from received headers that are not private. Also extract hostnames or domains.
        """
        received_items = self.email.get_all("received")
        if received_items is None:
            return
        for received in received_items:
            fromstr = re.split(r"\sby\s", received)[0].strip()
            if fromstr.startswith('from') is not True:
                continue
            for i in ['(', ')', '[', ']']:
                fromstr = fromstr.replace(i, " ")
            tokens = fromstr.split(" ")
            ip = None
            for token in tokens:
                try:
                    ip = ipaddress.ip_address(token)
                    break
                except ValueError:
                    pass  # token is not IP address

            if not ip or ip.is_private:
                continue  # skip header if IP not found or is private

            self.add_attribute("received-header-ip", value=str(ip), comment=fromstr)

        # The hostnames and/or domains always come after the "Received: from"
        # part so we can use regex to pick up those attributes.
        received_from = re.findall(r'(?<=from\s)[\w\d\.\-]+\.\w{2,24}', str(received_items))
        try:
            [self.add_attribute("received-header-hostname", i) for i in received_from]
        except Exception:
            pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from datetime import date, datetime

from deprecated import deprecated  # type: ignore
from json import JSONEncoder
from uuid import UUID
from abc import ABCMeta

try:
    from rapidjson import load  # type: ignore
    from rapidjson import loads  # type: ignore
    from rapidjson import dumps  # type: ignore
    HAS_RAPIDJSON = True
except ImportError:
    from json import load
    from json import loads
    from json import dumps
    HAS_RAPIDJSON = False

import logging
from enum import Enum
from typing import Union, Optional, Any, Dict, List, Set, Mapping

from .exceptions import PyMISPInvalidFormat, PyMISPError


from collections.abc import MutableMapping
from functools import lru_cache
from pathlib import Path

logger = logging.getLogger('pymisp')

resources_path = Path(__file__).parent / 'data'
misp_objects_path = resources_path / 'misp-objects' / 'objects'
with (resources_path / 'describeTypes.json').open('r') as f:
    describe_types = load(f)['result']


class MISPFileCache(object):
    # cache up to 150 JSON structures in class attribute

    @staticmethod
    @lru_cache(maxsize=150)
    def _load_json(path: Path) -> Union[dict, None]:
        if not path.exists():
            return None
        with path.open('r', encoding='utf-8') as f:
            data = load(f)
        return data


class Distribution(Enum):
    your_organisation_only = 0
    this_community_only = 1
    connected_communities = 2
    all_communities = 3
    sharing_group = 4
    inherit = 5


class ThreatLevel(Enum):
    high = 1
    medium = 2
    low = 3
    undefined = 4


class Analysis(Enum):
    initial = 0
    ongoing = 1
    completed = 2


def _int_to_str(d: Dict[str, Any]) -> Dict[str, Any]:
    # transform all integer back to string
    for k, v in d.items():
        if isinstance(v, dict):
            d[k] = _int_to_str(v)
        elif isinstance(v, int) and not isinstance(v, bool):
            d[k] = str(v)
    return d


@deprecated(reason=" Use method default=pymisp_json_default instead of cls=MISPEncode", version='2.4.117', action='default')
class MISPEncode(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, AbstractMISP):
            return obj.jsonable()
        elif isinstance(obj, (datetime, date)):
            return obj.isoformat()
        elif isinstance(obj, Enum):
            return obj.value
        elif isinstance(obj, UUID):
            return str(obj)
        return JSONEncoder.default(self, obj)


class AbstractMISP(MutableMapping, MISPFileCache, metaclass=ABCMeta):
    __resources_path = resources_path
    __misp_objects_path = misp_objects_path
    __describe_types = describe_types

    def __init__(self, **kwargs):
        """Abstract class for all the MISP objects.
        NOTE: Every method in every classes inheriting this one are doing
              changes in memory and  do not modify data on a remote MISP instance.
              To do so, you need to call the respective add_* or update_*
              methods in PyMISP.
        """
        super().__init__()
        self.__edited: bool = True  # As we create a new object, we assume it is edited
        self.__not_jsonable: List[str] = []
        self._fields_for_feed: Set
        self.__self_defined_describe_types: Optional[Dict] = None
        self.uuid: str

        if kwargs.get('force_timestamps') is not None:
            # Ignore the edited objects and keep the timestamps.
            self.__force_timestamps: bool = True
        else:
            self.__force_timestamps: bool = False

    @property
    def describe_types(self) -> Dict:
        if self.__self_defined_describe_types:
            return self.__self_defined_describe_types
        return self.__describe_types

    @describe_types.setter
    def describe_types(self, describe_types: Dict):
        self.__self_defined_describe_types = describe_types

    @property
    def resources_path(self) -> Path:
        return self.__resources_path

    @property
    def misp_objects_path(self) -> Path:
        return self.__misp_objects_path

    @misp_objects_path.setter
    def misp_objects_path(self, misp_objects_path: Union[str, Path]):
        if isinstance(misp_objects_path, str):
            misp_objects_path = Path(misp_objects_path)
        self.__misp_objects_path = misp_objects_path

    def from_dict(self, **kwargs) -> None:
        """Loading all the parameters as class properties, if they aren't `None`.
        This method aims to be called when all the properties requiring a special
        treatment are processed.
        Note: This method is used when you initialize an object with existing data so by default,
        the class is flaged as not edited."""
        for prop, value in kwargs.items():
            if value is None:
                continue
            setattr(self, prop, value)
        # We load an existing dictionary, marking it an not-edited
        self.__edited = False

    def update_not_jsonable(self, *args) -> None:
        """Add entries to the __not_jsonable list"""
        self.__not_jsonable += args

    def set_not_jsonable(self, args: List[str]) -> None:
        """Set __not_jsonable to a new list"""
        self.__not_jsonable = args

    def _remove_from_not_jsonable(self, *args) -> None:
        """Remove the entries that are in the __not_jsonable list"""
        for entry in args:
            try:
                self.__not_jsonable.remove(entry)
            except ValueError:
                pass

    def from_json(self, json_string: str) -> None:
        """Load a JSON string"""
        self.from_dict(**loads(json_string))

    def to_dict(self, json_format: bool = False) -> Dict:
        """Dump the class to a dictionary.
        This method automatically removes the timestamp recursively in every object
        that has been edited is order to let MISP update the event accordingly."""
        is_edited = self.edited
        to_return = {}
        for attribute, val in self.items():
            if val is None:
                continue
            elif isinstance(val, list) and len(val) == 0:
                continue
            elif isinstance(val, str):
                val = val.strip()
            elif json_format:
                if isinstance(val, AbstractMISP):
                    val = val.to_json(True)
                elif isinstance(val, (datetime, date)):
                    val = val.isoformat()
                elif isinstance(val, Enum):
                    val = val.value
                elif isinstance(val, UUID):
                    val = str(val)

            if attribute == 'timestamp':
                if not self.__force_timestamps and is_edited:
                    # In order to be accepted by MISP, the timestamp of an object
                    # needs to be either newer, or None.
                    # If the current object is marked as edited, the easiest is to
                    # skip the timestamp and let MISP deal with it
                    continue
                else:
                    val = self._datetime_to_timestamp(val)
            if (attribute in ('first_seen', 'last_seen', 'datetime')
                    and isinstance(val, datetime)
                    and not val.tzinfo):
                # Need to make sure the timezone is set. Otherwise, it will be processed as UTC on the server
                val = val.astimezone()

            to_return[attribute] = val
        to_return = _int_to_str(to_return)
        return to_return

    def jsonable(self) -> Dict:
        """This method is used by the JSON encoder"""
        return self.to_dict()

    def _to_feed(self) -> Dict:
        if not hasattr(self, '_fields_for_feed') or not self._fields_for_feed:
            raise PyMISPError('Unable to export in the feed format, _fields_for_feed is missing.')
        if hasattr(self, '_set_default') and callable(self._set_default):  # type: ignore
            self._set_default()  # type: ignore
        to_return = {}
        for field in sorted(self._fields_for_feed):
            if getattr(self, field, None) is not None:
                if field in ['timestamp', 'publish_timestamp']:
                    to_return[field] = self._datetime_to_timestamp(getattr(self, field))
                elif isinstance(getattr(self, field), (datetime, date)):
                    to_return[field] = getattr(self, field).isoformat()
                else:
                    to_return[field] = getattr(self, field)
            else:
                if field in ['data', 'first_seen', 'last_seen', 'deleted']:
                    # special fields
                    continue
                raise PyMISPError('The field {} is required in {} when generating a feed.'.format(field, self.__class__.__name__))
        to_return = _int_to_str(to_return)
        return to_return

    def to_json(self, sort_keys: bool = False, indent: Optional[int] = None) -> str:
        """Dump recursively any class of type MISPAbstract to a json string"""
        return dumps(self, default=pymisp_json_default, sort_keys=sort_keys, indent=indent)

    def __getitem__(self, key):
        try:
            if key[0] != '_' and key not in self.__not_jsonable:
                return self.__dict__[key]
            raise KeyError
        except AttributeError:
            # Expected by pop and other dict-related methods
            raise KeyError

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __delitem__(self, key):
        delattr(self, key)

    def __iter__(self):
        '''When we call **self, skip keys:
            * starting with _
            * in __not_jsonable
            * timestamp if the object is edited *unless* it is forced
        '''
        return iter({k: v for k, v in self.__dict__.items()
                     if not (k[0] == '_'
                             or k in self.__not_jsonable
                             or (not self.__force_timestamps and (k == 'timestamp' and self.__edited)))})

    def __len__(self) -> int:
        return len([k for k in self.__dict__.keys() if not (k[0] == '_' or k in self.__not_jsonable)])

    @property
    def force_timestamp(self) -> bool:
        return self.__force_timestamps

    @force_timestamp.setter
    def force_timestamp(self, force: bool):
        self.__force_timestamps = force

    @property
    def edited(self) -> bool:
        """Recursively check if an object has been edited and update the flag accordingly
        to the parent objects"""
        if self.__edited:
            return self.__edited
        for p, val in self.items():
            if isinstance(val, AbstractMISP) and val.edited:
                self.__edited = True
                break
            elif isinstance(val, list) and all(isinstance(a, AbstractMISP) for a in val):
                if any(a.edited for a in val):
                    self.__edited = True
                    break
        return self.__edited

    @edited.setter
    def edited(self, val: bool):
        """Set the edit flag"""
        if isinstance(val, bool):
            self.__edited = val
        else:
            raise PyMISPError('edited can only be True or False')

    def __setattr__(self, name: str, value: Any):
        if name[0] != '_' and not self.__edited and name in self:
            # The private members don't matter
            # If we already have a key with that name, we're modifying it.
            self.__edited = True
        super().__setattr__(name, value)

    def _datetime_to_timestamp(self, d: Union[int, float, str, datetime]) -> int:
        """Convert a datetime object to a timestamp (int)"""
        if isinstance(d, (int, float, str)):
            # Assume we already have a timestamp
            return int(d)
        return int(d.timestamp())

    def _add_tag(self, tag: Optional[Union[str, 'MISPTag', Mapping]] = None, **kwargs):
        """Add a tag to the attribute (by name or a MISPTag object)"""
        if isinstance(tag, str):
            misp_tag = MISPTag()
            misp_tag.from_dict(name=tag)
        elif isinstance(tag, MISPTag):
            misp_tag = tag
        elif isinstance(tag, dict):
            misp_tag = MISPTag()
            misp_tag.from_dict(**tag)
        elif kwargs:
            misp_tag = MISPTag()
            misp_tag.from_dict(**kwargs)
        else:
            raise PyMISPInvalidFormat(f"The tag is in an invalid format (can be either string, MISPTag, or an expanded dict): {tag}")
        if misp_tag not in self.tags:  # type: ignore
            self.Tag.append(misp_tag)
            self.edited = True
        return misp_tag

    def _set_tags(self, tags: List['MISPTag']):
        """Set a list of prepared MISPTag."""
        if all(isinstance(x, MISPTag) for x in tags):
            self.Tag = tags
        else:
            raise PyMISPInvalidFormat('All the attributes have to be of type MISPTag.')

    def __eq__(self, other) -> bool:
        if isinstance(other, AbstractMISP):
            return self.to_dict() == other.to_dict()
        elif isinstance(other, dict):
            return self.to_dict() == other
        else:
            return False

    def __repr__(self) -> str:
        return '<{self.__class__.__name__} - please define me>'.format(self=self)


class MISPTag(AbstractMISP):

    _fields_for_feed: set = {'name', 'colour'}

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name: str
        self.exportable: bool
        self.local: bool

    def from_dict(self, **kwargs):
        if kwargs.get('Tag'):
            kwargs = kwargs.get('Tag')
        super().from_dict(**kwargs)

    def _set_default(self):
        if not hasattr(self, 'colour'):
            self.colour = '#ffffff'

    def _to_feed(self, with_local: bool = True) -> Dict:
        if hasattr(self, 'exportable') and not self.exportable:
            return {}
        if with_local is False and hasattr(self, 'local') and self.local:
            return {}
        return super()._to_feed()

    def delete(self):
        self.deleted = True
        self.edited = True

    def __repr__(self) -> str:
        if hasattr(self, 'name'):
            return '<{self.__class__.__name__}(name={self.name})>'.format(self=self)
        return '<{self.__class__.__name__}(NotInitialized)>'.format(self=self)


if HAS_RAPIDJSON:
    def pymisp_json_default(obj: Union[AbstractMISP, datetime, date, Enum, UUID]) -> Union[Dict, str]:
        if isinstance(obj, AbstractMISP):
            return obj.jsonable()
        elif isinstance(obj, (datetime, date)):
            return obj.isoformat()
        elif isinstance(obj, Enum):
            return obj.value
        elif isinstance(obj, UUID):
            return str(obj)
else:
    def pymisp_json_default(obj: Union[AbstractMISP, datetime, date, Enum, UUID]) -> Union[Dict, str]:
        if isinstance(obj, AbstractMISP):
            return obj.jsonable()
        elif isinstance(obj, (datetime, date)):
            return obj.isoformat()
        elif isinstance(obj, Enum):
            return obj.value
        elif isinstance(obj, UUID):
            return str(obj)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .abstractgenerator import AbstractMISPObjectGenerator
import logging

logger = logging.getLogger('pymisp')


class Fail2BanObject(AbstractMISPObjectGenerator):

    def __init__(self, parameters: dict, strict: bool = True, **kwargs):
        super().__init__('fail2ban', strict=strict, **kwargs)
        self._parameters = parameters
        self.generate_attributes()

    def generate_attributes(self):
        timestamp = self._sanitize_timestamp(self._parameters.pop('processing-timestamp', None))
        self._parameters['processing-timestamp'] = timestamp
        super().generate_attributes()



# -*- coding: utf-8 -*-

try:
    from misp_stix_converter.converters.buildMISPAttribute import buildEvent  # type: ignore
    from misp_stix_converter.converters import convert  # type: ignore
    from misp_stix_converter.converters.convert import MISPtoSTIX  # type: ignore
    has_misp_stix_converter = True
except ImportError:
    has_misp_stix_converter = False


def load_stix(stix, distribution: int = 3, threat_level_id: int = 2, analysis: int = 0):
    '''Returns a MISPEvent object from a STIX package'''
    if not has_misp_stix_converter:
        raise Exception('You need to install misp_stix_converter: pip install git+https://github.com/MISP/MISP-STIX-Converter.git')
    stix = convert.load_stix(stix)
    return buildEvent(stix, distribution=distribution,
                      threat_level_id=threat_level_id, analysis=analysis)


def make_stix_package(misp_event, to_json: bool = False, to_xml: bool = False):
    '''Returns a STIXPackage from a MISPEvent.

       Optionally can return the package in json or xml.

    '''
    if not has_misp_stix_converter:
        raise Exception('You need to install misp_stix_converter: pip install git+https://github.com/MISP/MISP-STIX-Converter.git')
    package = MISPtoSTIX(misp_event)
    if to_json:
        return package.to_json()
    elif to_xml:
        return package.to_xml()
    else:
        return package



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from .abstractgenerator import AbstractMISPObjectGenerator


class SBSignatureObject(AbstractMISPObjectGenerator):
    '''
    Sandbox Analyzer
    '''
    def __init__(self, software: str, report: list, **kwargs):
        super().__init__('sb-signature', **kwargs)
        self._software = software
        self._report = report
        self.generate_attributes()

    def generate_attributes(self):
        ''' Parse the report for relevant attributes '''
        self.add_attribute("software", value=self._software)
        for (signature_name, description) in self._report:
            self.add_attribute("signature", value=signature_name, comment=description)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from ..exceptions import InvalidMISPObject
from .abstractgenerator import AbstractMISPObjectGenerator
from io import StringIO
import logging
from typing import Optional, Union
from pathlib import Path

logger = logging.getLogger('pymisp')


class SSHAuthorizedKeysObject(AbstractMISPObjectGenerator):

    def __init__(self, authorized_keys_path: Optional[Union[Path, str]] = None, authorized_keys_pseudofile: Optional[StringIO] = None, **kwargs):
        # PY3 way:
        super().__init__('ssh-authorized-keys', **kwargs)
        if authorized_keys_path:
            with open(authorized_keys_path, 'r') as f:
                self.__pseudofile = StringIO(f.read())
        elif authorized_keys_pseudofile and isinstance(authorized_keys_pseudofile, StringIO):
            self.__pseudofile = authorized_keys_pseudofile
        else:
            raise InvalidMISPObject('File buffer (StringIO) or a path is required.')
        self.__data = self.__pseudofile.getvalue()
        self.generate_attributes()

    def generate_attributes(self):
        for line in self.__pseudofile:
            if line.startswith('ssh') or line.startswith('ecdsa'):
                key = line.split(' ')[1]
                self.add_attribute('key', key)



# -*- coding: utf-8 -*-

import glob
import os
from .. import MISPEvent

try:
    from py2neo import authenticate, Graph, Node, Relationship  # type: ignore
    has_py2neo = True
except ImportError:
    has_py2neo = False


class Neo4j():

    def __init__(self, host='localhost:7474', username='neo4j', password='neo4j'):
        if not has_py2neo:
            raise Exception('py2neo is required, please install: pip install py2neo')
        authenticate(host, username, password)
        self.graph = Graph("http://{}/db/data/".format(host))

    def load_events_directory(self, directory):
        self.events = []
        for path in glob.glob(os.path.join(directory, '*.json')):
            e = MISPEvent()
            e.load(path)
            self.import_event(e)

    def del_all(self):
        self.graph.delete_all()

    def import_event(self, event):
        tx = self.graph.begin()
        event_node = Node('Event', uuid=event.uuid, name=event.info)
        # event_node['distribution'] = event.distribution
        # event_node['threat_level_id'] = event.threat_level_id
        # event_node['analysis'] = event.analysis
        # event_node['published'] = event.published
        # event_node['date'] = event.date.isoformat()
        tx.create(event_node)
        for a in event.attributes:
            attr_node = Node('Attribute', a.type, uuid=a.uuid)
            attr_node['category'] = a.category
            attr_node['name'] = a.value
            # attr_node['to_ids'] = a.to_ids
            # attr_node['comment'] = a.comment
            # attr_node['distribution'] = a.distribution
            tx.create(attr_node)
            member_rel = Relationship(event_node, "is member", attr_node)
            tx.create(member_rel)
            val = Node('Value', name=a.value)
            ev = Relationship(event_node, "has", val)
            av = Relationship(attr_node, "is", val)
            s = val | ev | av
            tx.merge(s)
            # tx.graph.push(s)
        tx.commit()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .. import MISPObject
from ..exceptions import InvalidMISPObject
from datetime import datetime, date
from dateutil.parser import parse
from typing import Union, Optional


class AbstractMISPObjectGenerator(MISPObject):

    def _detect_epoch(self, timestamp: Union[str, int, float]) -> bool:
        try:
            tmp = float(timestamp)
            if tmp < 30000000:
                # Assuming the user doesn't want to report anything before datetime(1970, 12, 14, 6, 20)
                # The date is most probably in the format 20180301
                return False
            return True
        except ValueError:
            return False

    def _sanitize_timestamp(self, timestamp: Optional[Union[datetime, date, dict, str, int, float]] = None) -> datetime:
        if not timestamp:
            return datetime.now()

        if isinstance(timestamp, datetime):
            return timestamp
        elif isinstance(timestamp, date):
            return datetime.combine(timestamp, datetime.min.time())
        elif isinstance(timestamp, dict):
            if not isinstance(timestamp['value'], datetime):
                timestamp['value'] = parse(timestamp['value'])
            return timestamp['value']
        else:  # Supported: float/int/string
            if isinstance(timestamp, (str, int, float)) and self._detect_epoch(timestamp):
                # It converts to the *local* datetime, which is consistent with the rest of the code.
                return datetime.fromtimestamp(float(timestamp))
            elif isinstance(timestamp, str):
                return parse(timestamp)
            else:
                raise Exception(f'Unable to convert {timestamp} to a datetime.')

    def generate_attributes(self):
        """Contains the logic where all the values of the object are gathered"""
        if hasattr(self, '_parameters'):
            for object_relation in self._definition['attributes']:
                value = self._parameters.pop(object_relation, None)
                if not value:
                    continue
                if isinstance(value, dict):
                    self.add_attribute(object_relation, **value)
                elif isinstance(value, list):
                    self.add_attributes(object_relation, *value)
                else:
                    # Assume it is the value only
                    self.add_attribute(object_relation, value=value)
            if self._strict and self._known_template and self._parameters:
                raise InvalidMISPObject('Some object relations are unknown in the template and could not be attached: {}'.format(', '.join(self._parameters)))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#
#    A simple converter of MISP objects to asciidoctor format
#    Copyright (C) 2017-2021 Alexandre Dulaunoy
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os
import json
import argparse

thisDir = os.path.dirname(__file__)

objects = []

pathObjects = os.path.join(thisDir, '../objects')

for f in os.listdir(pathObjects):
    objectName = f
    objects.append(objectName)

objects.sort()

argParser = argparse.ArgumentParser(description='Generate list of MISP object templates', epilog='Available objects are {0}'.format(objects))
argParser.add_argument('-v', action='store_true', help='Verbose mode')
args = argParser.parse_args()


for mispobject in objects:
    fullPathClusters = os.path.join(pathObjects, '{}/{}'.format(mispobject, 'definition.json'))
    with open(fullPathClusters) as fp:
        c = json.load(fp)
    if not c['description'].endswith('.'):
        c['description'] = c['description'] + "."
    v = "- [objects/{}](https://github.com/MISP/misp-objects/blob/main/objects/{}/definition.json) - {}".format(c['name'], c['name'], c['description'])
    print(v)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#
#    A simple converter of MISP objects to asciidoctor format
#    Copyright (C) 2017-2019 Alexandre Dulaunoy
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os
import json
import argparse

thisDir = os.path.dirname(__file__)

objects = []

pathObjects = os.path.join(thisDir, '../objects')

for f in os.listdir(pathObjects):
    objectName = f
    objects.append(objectName)

objects.sort()

argParser = argparse.ArgumentParser(description='Generate documentation from MISP objects', epilog='Available objects are {0}'.format(objects))
argParser.add_argument('-v', action='store_true', help='Verbose mode')
args = argParser.parse_args()


def header(adoc=False):
    if adoc is False:
        return False
    doc = adoc
    dedication = "\n[dedication]\n== Funding and Support\nThe MISP project is financially and resource supported by https://www.circl.lu/[CIRCL Computer Incident Response Center Luxembourg ].\n\nimage:{images-misp}logo.png[CIRCL logo]\n\nA CEF (Connecting Europe Facility) funding under CEF-TC-2016-3 - Cyber Security has been granted from 1st September 2017 until 31th August 2019 as ***Improving MISP as building blocks for next-generation information sharing***.\n\nimage:{images-misp}en_cef.png[CEF funding]\n\nIf you are interested to co-fund projects around MISP, feel free to get in touch with us.\n\n"
    doc = doc + ":toc: right\n"
    doc = doc + ":toclevels: 1\n"
    doc = doc + ":toc-title: MISP Objects\n"
    doc = doc + ":icons: font\n"
    doc = doc + ":sectanchors:\n"
    doc = doc + ":sectlinks:\n"
    doc = doc + ":images-cdn: https://raw.githubusercontent.com/MISP/MISP/2.4/INSTALL/logos/\n"
    doc = doc + ":images-misp: https://www.misp-project.org/assets/images/\n"
    doc = doc + "\n= MISP Objects\n\n"
    doc = doc + "= Introduction\n"
    doc = doc + "\nimage::{images-cdn}misp-logo.png[MISP logo]\n"
    doc = doc + "The MISP threat sharing platform is a free and open source software helping information sharing of threat intelligence including cyber security indicators, financial fraud or counter-terrorism information. The MISP project includes multiple sub-projects to support the operational requirements of analysts and improve the overall quality of information shared.\n\n"
    doc = doc + ""
    doc = "{}{}".format(doc, "\nMISP objects are used in MISP (starting from version 2.4.80) system and can be used by other information sharing tool.  MISP objects are in addition to MISP attributes to allow advanced combinations of attributes. The creation of these objects and their associated attributes are based on real cyber security use-cases and existing practices in information sharing. The objects are just shared like any other attributes in MISP even if the other MISP instances don't have the template of the object.\n")
    doc = doc + "The following document is generated from the machine-readable JSON describing the https://github.com/MISP/misp-objects[MISP objects]."
    doc = doc + "\n\n"
    doc = doc + "<<<\n"
    doc = doc + dedication
    doc = doc + "<<<\n"
    doc = doc + "= MISP objects\n"
    return doc


def asciidoc(content=False, adoc=None, t='title', title=''):

    adoc = adoc + "\n"
    output = ""
    if t == 'title':
        output = '== ' + content
    elif t == 'info':
        content = content.rstrip('.')
        output = "\n{}.\n\n{} {} {}{}/definition.json[*this location*] {}.\n".format(content, 'NOTE: ', title, 'is a MISP object available in JSON format at https://github.com/MISP/misp-objects/blob/main/objects/', title.lower(), ' The JSON format can be freely reused in your application or automatically enabled in https://www.github.com/MISP/MISP[MISP]')
    elif t == 'author':
        output = '\nauthors:: {}\n'.format(' - '.join(content))
    elif t == 'value':
        output = '=== ' + content
    elif t == 'description':
        content = content.rstrip('.')
        output = '\n{}\n'.format(content)
    elif t == 'attributes':
        # output = '\n{}\n'.format
        # output = '[cols=\",a\"]\n'
        output = output + '|===\n'
        output = output + '|Object attribute | MISP attribute type | Description | Disable correlation | Multiple\n'
        adoc = adoc + output
        for v in content['attributes']:
            disableCorrelation = 'icon:minus[] '
            description = 'icon:minus[] '
            multiple = 'icon:minus[] '
            if 'disable_correlation' in content['attributes'][v]:
                if content['attributes'][v]['disable_correlation']:
                    disableCorrelation = 'icon:check[] '
            if 'multiple' in content['attributes'][v]:
                if content['attributes'][v]['multiple']:
                    multiple = 'icon:check[] '
            if 'description' in content['attributes'][v]:
                if content['attributes'][v]['description']:
                    description = '{}'.format(content['attributes'][v]['description'])
                if 'values_list' in content['attributes'][v]:
                    values = content['attributes'][v]['values_list']
                    description = '{} {}'.format(content['attributes'][v]['description'], values)
                if 'sane_default' in content['attributes'][v]:
                    values = content['attributes'][v]['sane_default']
                    description = '{} {}'.format(content['attributes'][v]['description'], values)
            output = '\n| {} | {} a| {} a| {} a| {}\n'.format(v, content['attributes'][v]['misp-attribute'], description, disableCorrelation, multiple)
            adoc = adoc + output
        output = '\n|===\n'
    adoc = adoc + output
    return adoc


adoc = ""
print(header(adoc=adoc))

for mispobject in objects:
    fullPathClusters = os.path.join(pathObjects, '{}/{}'.format(mispobject, 'definition.json'))
    with open(fullPathClusters) as fp:
        c = json.load(fp)
    title = c['name']
    adoc = asciidoc(content=title, adoc=adoc, t='title')
    adoc = asciidoc(content=c['description'], adoc=adoc, t='info', title=title)
    adoc = asciidoc(content=c, adoc=adoc, t='attributes', title=title)

with open('../relationships/definition.json') as filerelationships:
    rel = json.load(filerelationships)

    output = '== Relationships\n'
    output = output + '\n{}\n'.format(rel['description'])
    output = output + '\nRelationships are part of MISP object and available in JSON format at https://github.com/MISP/misp-objects/blob/main/relationships/definition.json[this location]. The JSON format can be freely reused in your application or automatically enabled in https://www.github.com/MISP/MISP[MISP].\n'
    output = output + '|===\n'
    output = output + '|Name of relationship | Description | Format\n'
    for relationship in rel['values']:
        output = output + '\n| {} | {} | {}\n'.format(relationship['name'], relationship['description'], str(relationship['format']))
    output = output + '\n|===\n'
    adoc = adoc + output


print(adoc)



# -*- coding: utf-8 -*-


class PyMISPError(Exception):
    def __init__(self, message):
        super(PyMISPError, self).__init__(message)
        self.message = message


class NewEventError(PyMISPError):
    pass


class UpdateEventError(PyMISPError):
    pass


class NewAttributeError(PyMISPError):
    pass


class NewEventReportError(PyMISPError):
    pass


class UpdateAttributeError(PyMISPError):
    pass


class NewGalaxyClusterError(PyMISPError):
    pass


class NewGalaxyClusterRelationError(PyMISPError):
    pass


class SearchError(PyMISPError):
    pass


class MissingDependency(PyMISPError):
    pass


class NoURL(PyMISPError):
    pass


class NoKey(PyMISPError):
    pass


class MISPObjectException(PyMISPError):
    pass


class InvalidMISPObject(MISPObjectException):
    """Exception raised when an object doesn't respect the contrains in the definition"""
    pass


class UnknownMISPObjectTemplate(MISPObjectException):
    """Exception raised when the template is unknown"""
    pass


class PyMISPInvalidFormat(PyMISPError):
    pass


class MISPServerError(PyMISPError):
    pass


class PyMISPNotImplementedYet(PyMISPError):
    pass


class PyMISPUnexpectedResponse(PyMISPError):
    pass


class PyMISPEmptyResponse(PyMISPError):
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from ..exceptions import InvalidMISPObject
from .abstractgenerator import AbstractMISPObjectGenerator
import os
from io import BytesIO
from hashlib import md5, sha1, sha256, sha512
import math
from collections import Counter
import logging
from pathlib import Path
from typing import Union

logger = logging.getLogger('pymisp')


try:
    import pydeep  # type: ignore
    HAS_PYDEEP = True
except ImportError:
    HAS_PYDEEP = False

try:
    import magic  # type: ignore
    HAS_MAGIC = True
except ImportError:
    HAS_MAGIC = False


class FileObject(AbstractMISPObjectGenerator):

    def __init__(self, filepath: Union[Path, str] = None, pseudofile: BytesIO = None, filename: str = None, **kwargs):
        super().__init__('file', **kwargs)
        if not HAS_PYDEEP:
            logger.warning("Please install pydeep: pip install git+https://github.com/kbandla/pydeep.git")
        if not HAS_MAGIC:
            logger.warning("Please install python-magic: pip install python-magic.")
        if filename:
            # Useful in case the file is copied with a pre-defined name by a script but we want to keep the original name
            self.__filename = filename
        elif filepath:
            self.__filename = os.path.basename(filepath)
        else:
            raise InvalidMISPObject('A file name is required (either in the path, or as a parameter).')

        if filepath:
            with open(filepath, 'rb') as f:
                self.__pseudofile = BytesIO(f.read())
        elif pseudofile and isinstance(pseudofile, BytesIO):
            # WARNING: lief.parse requires a path
            self.__pseudofile = pseudofile
        else:
            raise InvalidMISPObject('File buffer (BytesIO) or a path is required.')
        self.__data = self.__pseudofile.getvalue()
        self.generate_attributes()

    def generate_attributes(self):
        self.add_attribute('filename', value=self.__filename)
        size = self.add_attribute('size-in-bytes', value=len(self.__data))
        if int(size.value) > 0:
            self.add_attribute('entropy', value=self.__entropy_H(self.__data))
            self.add_attribute('md5', value=md5(self.__data).hexdigest())
            self.add_attribute('sha1', value=sha1(self.__data).hexdigest())
            self.add_attribute('sha256', value=sha256(self.__data).hexdigest())
            self.add_attribute('sha512', value=sha512(self.__data).hexdigest())
            self.add_attribute('malware-sample', value=self.__filename, data=self.__pseudofile, disable_correlation=True)
            if HAS_MAGIC:
                self.add_attribute('mimetype', value=magic.from_buffer(self.__data, mime=True))
            if HAS_PYDEEP:
                self.add_attribute('ssdeep', value=pydeep.hash_buf(self.__data).decode())

    def __entropy_H(self, data: bytes) -> float:
        """Calculate the entropy of a chunk of data."""
        # NOTE: copy of the entropy function from pefile

        if len(data) == 0:
            return 0.0

        occurrences = Counter(bytearray(data))

        entropy = 0.0
        for x in occurrences.values():
            p_x = float(x) / len(data)
            entropy -= p_x * math.log(p_x, 2)

        return entropy



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from ..exceptions import InvalidMISPObject
from .abstractgenerator import AbstractMISPObjectGenerator
from io import BytesIO
from hashlib import md5, sha1, sha256, sha512
from datetime import datetime
import logging
from typing import Optional, Union
from pathlib import Path
from base64 import b64encode

from . import FileObject

import lief  # type: ignore

try:
    import pydeep  # type: ignore
    HAS_PYDEEP = True
except ImportError:
    HAS_PYDEEP = False

logger = logging.getLogger('pymisp')


def make_pe_objects(lief_parsed: lief.Binary, misp_file: FileObject, standalone: bool = True, default_attributes_parameters: dict = {}):
    pe_object = PEObject(parsed=lief_parsed, standalone=standalone, default_attributes_parameters=default_attributes_parameters)
    misp_file.add_reference(pe_object.uuid, 'includes', 'PE indicators')
    pe_sections = []
    for s in pe_object.sections:
        pe_sections.append(s)
    return misp_file, pe_object, pe_sections


class PEObject(AbstractMISPObjectGenerator):

    def __init__(self, parsed: Optional[lief.PE.Binary] = None, filepath: Optional[Union[Path, str]] = None, pseudofile: Optional[BytesIO] = None, **kwargs):
        """Creates an PE object, with lief"""
        super().__init__('pe', **kwargs)
        if not HAS_PYDEEP:
            logger.warning("Please install pydeep: pip install git+https://github.com/kbandla/pydeep.git")
        if pseudofile:
            if isinstance(pseudofile, BytesIO):
                self.__pe = lief.PE.parse(raw=pseudofile.getvalue())
            elif isinstance(pseudofile, bytes):
                self.__pe = lief.PE.parse(raw=pseudofile)
            else:
                raise InvalidMISPObject('Pseudo file can be BytesIO or bytes got {}'.format(type(pseudofile)))
        elif filepath:
            self.__pe = lief.PE.parse(filepath)
        elif parsed:
            # Got an already parsed blob
            if isinstance(parsed, lief.PE.Binary):
                self.__pe = parsed
            else:
                raise InvalidMISPObject('Not a lief.PE.Binary: {}'.format(type(parsed)))
        self.generate_attributes()

    def _is_exe(self):
        if not self._is_dll() and not self._is_driver():
            return self.__pe.header.has_characteristic(lief.PE.HEADER_CHARACTERISTICS.EXECUTABLE_IMAGE)
        return False

    def _is_dll(self):
        return self.__pe.header.has_characteristic(lief.PE.HEADER_CHARACTERISTICS.DLL)

    def _is_driver(self):
        # List from pefile
        system_DLLs = set(('ntoskrnl.exe', 'hal.dll', 'ndis.sys', 'bootvid.dll', 'kdcom.dll'))
        if system_DLLs.intersection([imp.lower() for imp in self.__pe.libraries]):
            return True
        return False

    def _get_pe_type(self):
        if self._is_dll():
            return 'dll'
        elif self._is_driver():
            return 'driver'
        elif self._is_exe():
            return 'exe'
        else:
            return 'unknown'

    def generate_attributes(self):
        self.add_attribute('type', value=self._get_pe_type())
        # General information
        self.add_attribute('entrypoint-address', value=self.__pe.entrypoint)
        self.add_attribute('compilation-timestamp', value=datetime.utcfromtimestamp(self.__pe.header.time_date_stamps).isoformat())
        self.add_attribute('imphash', value=lief.PE.get_imphash(self.__pe, lief.PE.IMPHASH_MODE.PEFILE))
        self.add_attribute('authentihash', value=self.__pe.authentihash_sha256.hex())
        try:
            if (self.__pe.has_resources
                    and self.__pe.resources_manager.has_version
                    and self.__pe.resources_manager.version.has_string_file_info
                    and self.__pe.resources_manager.version.string_file_info.langcode_items):
                fileinfo = dict(self.__pe.resources_manager.version.string_file_info.langcode_items[0].items.items())
                self.add_attribute('original-filename', value=fileinfo.get('OriginalFilename'))
                self.add_attribute('internal-filename', value=fileinfo.get('InternalName'))
                self.add_attribute('file-description', value=fileinfo.get('FileDescription'))
                self.add_attribute('file-version', value=fileinfo.get('FileVersion'))
                self.add_attribute('lang-id', value=self.__pe.resources_manager.version.string_file_info.langcode_items[0].key)
                self.add_attribute('product-name', value=fileinfo.get('ProductName'))
                self.add_attribute('product-version', value=fileinfo.get('ProductVersion'))
                self.add_attribute('company-name', value=fileinfo.get('CompanyName'))
                self.add_attribute('legal-copyright', value=fileinfo.get('LegalCopyright'))
        except lief.read_out_of_bound:
            # The file is corrupted
            pass
        # Sections
        self.sections = []
        if self.__pe.sections:
            pos = 0
            for section in self.__pe.sections:
                if not section.name and not section.size:
                    # Skip section if name is none AND size is 0.
                    continue
                s = PESectionObject(section, standalone=self._standalone, default_attributes_parameters=self._default_attributes_parameters)
                self.add_reference(s.uuid, 'includes', 'Section {} of PE'.format(pos))
                if ((self.__pe.entrypoint >= section.virtual_address)
                        and (self.__pe.entrypoint < (section.virtual_address + section.virtual_size))):
                    self.add_attribute('entrypoint-section-at-position', value='{}|{}'.format(section.name, pos))
                pos += 1
                self.sections.append(s)
        self.add_attribute('number-sections', value=len(self.sections))
        # Signatures
        self.certificates = []
        self.signers = []
        for sign in self.__pe.signatures:
            for c in sign.certificates:
                cert_obj = PECertificate(c)
                self.add_reference(cert_obj.uuid, 'signed-by')
                self.certificates.append(cert_obj)
            for s_info in sign.signers:
                signer_obj = PESigners(s_info)
                self.add_reference(signer_obj.uuid, 'signed-by')
                self.signers.append(signer_obj)


class PECertificate(AbstractMISPObjectGenerator):

    def __init__(self, certificate: lief.PE.x509, **kwargs):
        super().__init__('x509')
        self.__certificate = certificate
        self.generate_attributes()

    def generate_attributes(self):
        self.add_attribute('issuer', value=self.__certificate.issuer)
        self.add_attribute('serial-number', value=self.__certificate.serial_number)
        self.add_attribute('validity-not-before', value=datetime(*self.__certificate.valid_from))
        self.add_attribute('validity-not-after', value=datetime(*self.__certificate.valid_to))
        self.add_attribute('version', value=self.__certificate.version)
        self.add_attribute('subject', value=self.__certificate.subject)
        self.add_attribute('signature_algorithm', value=self.__certificate.signature_algorithm)
        self.add_attribute('raw-base64', value=b64encode(self.__certificate.raw))


class PESigners(AbstractMISPObjectGenerator):

    def __init__(self, signer: lief.PE.SignerInfo, **kwargs):
        super().__init__('authenticode-signerinfo')
        self.__signer = signer
        self.generate_attributes()

    def generate_attributes(self):
        self.add_attribute('issuer', value=self.__signer.issuer)
        self.add_attribute('serial-number', value=self.__signer.serial_number)
        self.add_attribute('version', value=self.__signer.version)
        self.add_attribute('digest_algorithm', value=self.__signer.digest_algorithm.name)
        self.add_attribute('encryption_algorithm', value=self.__signer.encryption_algorithm.name)
        self.add_attribute('digest-base64', value=b64encode(self.__signer.encrypted_digest))
        info = self.__signer.get_attribute(lief.PE.SIG_ATTRIBUTE_TYPES.SPC_SP_OPUS_INFO)
        if info:
            self.add_attribute('program-name', value=info.program_name)
            self.add_attribute('url', value=info.more_info)


class PESectionObject(AbstractMISPObjectGenerator):

    def __init__(self, section: lief.PE.Section, **kwargs):
        """Creates an PE Section object. Object generated by PEObject."""
        super().__init__('pe-section')
        self.__section = section
        self.__data = bytes(self.__section.content)
        self.generate_attributes()

    def generate_attributes(self):
        self.add_attribute('name', value=self.__section.name)
        size = self.add_attribute('size-in-bytes', value=self.__section.size)
        if int(size.value) > 0:
            # zero-filled sections can create too many correlations
            to_ids = float(self.__section.entropy) > 0
            disable_correlation = not to_ids
            self.add_attribute('entropy', value=self.__section.entropy)
            self.add_attribute('md5', value=md5(self.__data).hexdigest(), disable_correlation=disable_correlation, to_ids=to_ids)
            self.add_attribute('sha1', value=sha1(self.__data).hexdigest(), disable_correlation=disable_correlation, to_ids=to_ids)
            self.add_attribute('sha256', value=sha256(self.__data).hexdigest(), disable_correlation=disable_correlation, to_ids=to_ids)
            self.add_attribute('sha512', value=sha512(self.__data).hexdigest(), disable_correlation=disable_correlation, to_ids=to_ids)
            if HAS_PYDEEP and float(self.__section.entropy) > 0:
                if self.__section.name == '.rsrc':
                    # ssdeep of .rsrc creates too many correlations
                    disable_correlation = True
                    to_ids = False
                self.add_attribute('ssdeep', value=pydeep.hash_buf(self.__data).decode(), disable_correlation=disable_correlation, to_ids=to_ids)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from pathlib import Path
from pymisp import MISPEvent
import json
from typing import List


def feed_meta_generator(path: Path):
    manifests = {}
    hashes: List[str] = []

    for f_name in path.glob('*.json'):
        if str(f_name.name) == 'manifest.json':
            continue
        event = MISPEvent()
        event.load_file(str(f_name))
        manifests.update(event.manifest)
        hashes += [f'{h},{event.uuid}' for h in event.attributes_hashes('md5')]

    with (path / 'manifest.json').open('w') as f:
        json.dump(manifests, f)

    with (path / 'hashes.csv').open('w') as f:
        for h in hashes:
            f.write(f'{h}\n')
